"""Objects that define how data is recorded from simulation."""
from abc import ABC, abstractmethod
from typing import Union, Tuple

import pydantic
import numpy as np

from .types import Ax, EMField, ArrayLike, Bound, FreqArray
from .types import Literal, Direction, Coordinate, Axis, ObsGridArray, RadVec
from .geometry import Box
from .medium import Medium
from .validators import assert_plane
from .base import cached_property
from .mode import ModeSpec
from .viz import PlotParams, plot_params_monitor, ARROW_COLOR_MONITOR, ARROW_ALPHA
from ..log import SetupError, log, DataError, ValidationError
from ..constants import HERTZ, SECOND, MICROMETER, RADIAN


BYTES_REAL = 4
BYTES_COMPLEX = 8


class Monitor(Box, ABC):
    """Abstract base class for monitors."""

    name: str = pydantic.Field(
        ...,
        title="Name",
        description="Unique name for monitor.",
        min_length=1,
    )

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a Monitor object."""
        return plot_params_monitor

    @cached_property
    def geometry(self) -> Box:
        """:class:`Box` representation of monitor.

        Returns
        -------
        :class:`Box`
            Representation of the monitor geometry as a :class:`Box`.
        """
        return Box(center=self.center, size=self.size)

    @abstractmethod
    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization.

        Parameters
        ----------
        num_cells : int
            Number of grid cells within the monitor after discretization by a :class:`Simulation`.
        tmesh : Array
            The discretized time mesh of a :class:`Simulation`.

        Returns
        -------
        int
            Number of bytes to be stored in monitor.
        """


class FreqMonitor(Monitor, ABC):
    """:class:`Monitor` that records data in the frequency-domain."""

    freqs: FreqArray = pydantic.Field(
        ...,
        title="Frequencies",
        description="Array or list of frequencies stored by the field monitor.",
        units=HERTZ,
    )

    @pydantic.validator("freqs", always=True)
    def _freqs_non_empty(cls, val):
        """Assert one frequency present."""
        if len(val) == 0:
            raise ValidationError("'freqs' must not be empty.")
        return val


class TimeMonitor(Monitor, ABC):
    """:class:`Monitor` that records data in the time-domain."""

    start: pydantic.NonNegativeFloat = pydantic.Field(
        0.0,
        title="Start time",
        description="Time at which to start monitor recording.",
        units=SECOND,
    )

    stop: pydantic.NonNegativeFloat = pydantic.Field(
        None,
        title="Stop time",
        description="Time at which to stop monitor recording.  "
        "If not specified, record until end of simulation.",
        units=SECOND,
    )

    interval: pydantic.PositiveInt = pydantic.Field(
        1,
        title="Time interval",
        description="Number of time step intervals between monitor recordings.",
    )

    @pydantic.validator("stop", always=True, allow_reuse=True)
    def stop_greater_than_start(cls, val, values):
        """Ensure sure stop is greater than or equal to start."""
        start = values.get("start")
        if val and val < start:
            raise SetupError("Monitor start time is greater than stop time.")
        return val

    def time_inds(self, tmesh: ArrayLike[float, 1]) -> Tuple[int, int]:
        """Compute the starting and stopping index of the monitor in a given discrete time mesh."""

        tind_beg, tind_end = (0, 0)

        if tmesh.size == 0:
            return (tind_beg, tind_end)

        # If monitor.stop is None, record until the end
        t_stop = self.stop
        if t_stop is None:
            tind_end = int(tmesh.size)
            t_stop = tmesh[-1]
        else:
            tend = np.nonzero(tmesh <= t_stop)[0]
            if tend.size > 0:
                tind_end = int(tend[-1] + 1)

        # Step to compare to in order to handle t_start = t_stop
        dt = 1e-20 if np.array(tmesh).size < 2 else tmesh[1] - tmesh[0]
        # If equal start and stopping time, record one time step
        if np.abs(self.start - t_stop) < dt:
            tind_beg = max(tind_end - 1, 0)
        else:
            tbeg = np.nonzero(tmesh[:tind_end] >= self.start)[0]
            tind_beg = tbeg[0] if tbeg.size > 0 else tind_end
        return (tind_beg, tind_end)

    def num_steps(self, tmesh: ArrayLike[float, 1]) -> int:
        """Compute number of time steps for a time monitor."""

        tind_beg, tind_end = self.time_inds(tmesh)
        return int((tind_end - tind_beg) / self.interval)


class AbstractFieldMonitor(Monitor, ABC):
    """:class:`Monitor` that records electromagnetic field data as a function of x,y,z."""

    fields: Tuple[EMField, ...] = pydantic.Field(
        ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"],
        title="Field Components",
        description="Collection of field components to store in the monitor.",
    )

    interval_space: Tuple[
        pydantic.PositiveInt, pydantic.PositiveInt, pydantic.PositiveInt
    ] = pydantic.Field(
        (1, 1, 1),
        title="Spatial interval",
        description="Number of grid step intervals between monitor recordings. If equal to 1, "
        "there will be no downsampling. If greater than 1, fields will be downsampled "
        "and automatically colocated.",
    )

    colocate: bool = pydantic.Field(
        None,
        title="Colocate fields",
        description="Toggle whether fields should be colocated to grid cell centers. Default: "
        "``False`` if ``interval_space`` is 1 in each direction, ``True`` if ``interval_space`` "
        "is greater than one in any direction.",
    )

    @pydantic.validator("colocate", always=True)
    def set_default_colocate(cls, val, values):
        """Toggle default field colocation setting based on `interval_space`."""
        interval_space = values.get("interval_space")
        if val is None:
            val = sum(interval_space) != 3
        return val

    def downsampled_num_cells(self, num_cells: Tuple[int, int, int]) -> Tuple[int, int, int]:
        """Given a tuple of the number of cells spanned by the monitor along each dimension,
        return the number of cells one would have after downsampling based on ``interval_space``.
        """
        num_cells_new = list(num_cells.copy())
        for idx, interval in enumerate(self.interval_space):
            if interval == 1 or num_cells[idx] < 4 or (num_cells[idx] - 1) <= interval:
                continue
            num_cells_new[idx] = np.floor((num_cells[idx] - 1) / interval) + 1
            num_cells_new[idx] += int((num_cells[idx] - 1) % interval > 0)
        return num_cells_new


class PlanarMonitor(Monitor, ABC):
    """:class:`Monitor` that has a planar geometry."""

    _plane_validator = assert_plane()


class AbstractModeMonitor(PlanarMonitor, FreqMonitor):
    """:class:`Monitor` that records mode-related data."""

    mode_spec: ModeSpec = pydantic.Field(
        ...,
        title="Mode Specification",
        description="Parameters to feed to mode solver which determine modes measured by monitor.",
    )

    def plot(  # pylint:disable=too-many-arguments
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        sim_bounds: Bound = None,
        **patch_kwargs,
    ) -> Ax:

        # call the monitor.plot() function first
        ax = super().plot(x=x, y=y, z=z, ax=ax, **patch_kwargs)

        kwargs_alpha = patch_kwargs.get("alpha")
        arrow_alpha = ARROW_ALPHA if kwargs_alpha is None else kwargs_alpha

        # and then add an arrow using the direction comuputed from `_dir_arrow`.
        ax = self._plot_arrow(
            x=x,
            y=y,
            z=z,
            ax=ax,
            direction=self._dir_arrow,
            color=ARROW_COLOR_MONITOR,
            alpha=arrow_alpha,
            both_dirs=True,
            sim_bounds=sim_bounds,
        )
        return ax

    @cached_property
    def _dir_arrow(self) -> Tuple[float, float, float]:
        """Source direction normal vector in cartesian coordinates."""
        dx = np.cos(self.mode_spec.angle_phi) * np.sin(self.mode_spec.angle_theta)
        dy = np.sin(self.mode_spec.angle_phi) * np.sin(self.mode_spec.angle_theta)
        dz = np.cos(self.mode_spec.angle_theta)
        return self.unpop_axis(dz, (dx, dy), axis=self.size.index(0.0))


class FieldMonitor(AbstractFieldMonitor, FreqMonitor):
    """:class:`Monitor` that records electromagnetic fields in the frequency domain.

    Example
    -------
    >>> monitor = FieldMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     fields=['Hx'],
    ...     freqs=[250e12, 300e12],
    ...     name='steady_state_monitor')
    """

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 complex number per grid cell, per frequency, per field
        return BYTES_COMPLEX * num_cells * len(self.freqs) * len(self.fields)


class FieldTimeMonitor(AbstractFieldMonitor, TimeMonitor):
    """:class:`Monitor` that records electromagnetic fields in the time domain.

    Example
    -------
    >>> monitor = FieldTimeMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     fields=['Hx'],
    ...     start=1e-13,
    ...     stop=5e-13,
    ...     interval=2,
    ...     name='movie_monitor')
    """

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 real number per grid cell, per time step, per field
        num_steps = self.num_steps(tmesh)
        return BYTES_REAL * num_steps * num_cells * len(self.fields)


class PermittivityMonitor(FreqMonitor):
    """:class:`Monitor` that records the diagonal components of the complex-valued relative
    permittivity tensor in the frequency domain. The recorded data has the same shape as a
    :class:`.FieldMonitor` of the same geometry: the permittivity values are saved at the
    Yee grid locations, and can be interpolated to any point inside the monitor.

    Example
    -------
    >>> monitor = PermittivityMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     freqs=[250e12, 300e12],
    ...     name='eps_monitor')
    """

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 3 complex number per grid cell, per frequency
        return BYTES_COMPLEX * num_cells * len(self.freqs) * 3


class SurfaceIntegrationMonitor(Monitor, ABC):
    """Abstract class for monitors that perform surface integrals during the solver run, as in
    flux and near to far transformations."""

    normal_dir: Direction = pydantic.Field(
        None,
        title="Normal vector orientation",
        description="Direction of the surface monitor's normal vector w.r.t. "
        "the positive x, y or z unit vectors. Must be one of ``'+'`` or ``'-'``. "
        "Applies to surface monitors only, and defaults to ``'+'`` if not provided.",
    )

    exclude_surfaces: Tuple[Literal["x-", "x+", "y-", "y+", "z-", "z+"], ...] = pydantic.Field(
        None,
        title="Excluded surfaces",
        description="Surfaces to exclude in the integration, if a volume monitor.",
    )

    @pydantic.root_validator(skip_on_failure=True)
    def normal_dir_exists_for_surface(cls, values):
        """If the monitor is a surface, set default ``normal_dir`` if not provided.
        If the monitor is a box, warn that ``normal_dir`` is relevant only for surfaces."""
        normal_dir = values.get("normal_dir")
        name = values.get("name")
        size = values.get("size")
        if size.count(0.0) != 1:
            if normal_dir is not None:
                log.warning(
                    "The ``normal_dir`` field is relevant only for surface monitors "
                    f"and will be ignored for monitor {name}, which is a box."
                )
        else:
            if normal_dir is None:
                values["normal_dir"] = "+"
        return values

    @pydantic.root_validator(skip_on_failure=True)
    def check_excluded_surfaces(cls, values):
        """Error if ``exclude_surfaces`` is provided for a surface monitor."""
        exclude_surfaces = values.get("exclude_surfaces")
        if exclude_surfaces is None:
            return values
        name = values.get("name")
        size = values.get("size")
        if size.count(0.0) > 0:
            raise SetupError(
                f"Can't specify ``exclude_surfaces`` for surface monitor {name}; "
                "valid for box monitors only."
            )
        return values


class AbstractFluxMonitor(SurfaceIntegrationMonitor, ABC):
    """:class:`Monitor` that records flux during the solver run."""


class FluxMonitor(AbstractFluxMonitor, FreqMonitor):
    """:class:`Monitor` that records power flux in the frequency domain.
    If the monitor geometry is a 2D box, the total flux through this plane is returned, with a
    positive sign corresponding to power flow in the positive direction along the axis normal to
    the plane. If the geometry is a 3D box, the returned array has a ``surface`` coordinate, and
    stores the flux through each of the six surfaces of the box. The sign convention in that case
    is such that positive sign corresponds to power flowing outside of the box, such that
    summing over the ``surface`` dimension results in the total out-going power.

    Example
    -------
    >>> monitor = FluxMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     freqs=[200e12, 210e12],
    ...     name='flux_monitor')

    Note
    ----
    For a 2D plane, the flux is summed up over all Yee grid pixels that are touched by the plane,
    rather than integrated over the exact span of the plane. For a 3D monitor, this is also the
    case, but care is taken to not over- or under-count the power at the edges. Because of this,
    there can be small discrepancies between using a 3D FluxMonitor and manually placing six
    2D monitors at the surface locations.
    """

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 real number per frequency
        return BYTES_REAL * len(self.freqs)


class FluxTimeMonitor(AbstractFluxMonitor, TimeMonitor):
    """:class:`Monitor` that records power flux in the time domain.
    If the monitor geometry is a 2D box, the total flux through this plane is returned, with a
    positive sign corresponding to power flow in the positive direction along the axis normal to
    the plane. If the geometry is a 3D box, the returned array has a ``surface`` coordinate, and
    stores the flux through each of the six surfaces of the box. The sign convention in that case
    is such that positive sign corresponds to power flowing outside of the box, such that
    summing over the ``surface`` dimension results in the total out-going power.

    Example
    -------
    >>> monitor = FluxTimeMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     start=1e-13,
    ...     stop=5e-13,
    ...     interval=2,
    ...     name='flux_vs_time')

    Note
    ----
    For a 2D plane, the flux is summed up over all Yee grid pixels that are touched by the plane,
    rather than integrated over the exact span of the plane. For a 3D monitor, this is also the
    case, but care is taken to not over- or under-count the power at the edges. Because of this,
    there can be small discrepancies between using a 3D FluxMonitor and manually placing six
    2D monitors at the surface locations.
    """

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 real number per time tep
        num_steps = self.num_steps(tmesh)
        return BYTES_REAL * num_steps


class ModeMonitor(AbstractModeMonitor):
    """:class:`Monitor` that records amplitudes from modal decomposition of fields on plane.

    Example
    -------
    >>> mode_spec = ModeSpec(num_modes=3)
    >>> monitor = ModeMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     freqs=[200e12, 210e12],
    ...     mode_spec=mode_spec,
    ...     name='mode_monitor')
    """

    def storage_size(self, num_cells: int, tmesh: int) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 3 complex numbers per frequency, per mode.
        return 3 * BYTES_COMPLEX * len(self.freqs) * self.mode_spec.num_modes


class ModeSolverMonitor(AbstractModeMonitor):
    """:class:`Monitor` that stores the mode field profiles returned by the mode solver in the
    monitor plane.

    Example
    -------
    >>> mode_spec = ModeSpec(num_modes=3)
    >>> monitor = ModeSolverMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     freqs=[200e12, 210e12],
    ...     mode_spec=mode_spec,
    ...     name='mode_monitor')
    """

    def storage_size(self, num_cells: int, tmesh: int) -> int:
        """Size of monitor storage given the number of points after discretization."""
        return 6 * BYTES_COMPLEX * num_cells * len(self.freqs) * self.mode_spec.num_modes


class AbstractNear2FarMonitor(SurfaceIntegrationMonitor, FreqMonitor):
    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
    and invokes the computation of far fields.
    """

    fields: Tuple[RadVec, ...] = pydantic.Field(
        ["Ntheta", "Nphi", "Ltheta", "Lphi"],
        title="Field Components",
        description="Collection of radiation vector components to store in the monitor.",
    )

    custom_origin: Coordinate = pydantic.Field(
        None,
        title="Local origin",
        description="Local origin used for defining observation points. If ``None``, uses the "
        "monitor's center.",
        units=MICROMETER,
    )

    medium: Medium = pydantic.Field(
        Medium(permittivity=1),
        title="Background medium",
        description="Background medium in which to radiate near fields to far fields. "
        "If not provided, uses free space.",
    )

    @property
    def axis(self) -> Axis:
        """Returns the :class:`.Axis` normal to this surface."""
        # assume that the monitor's axis is in the direction where the monitor is thinnest
        # and errors if the monitor is a box
        if self.size.count(0.0) != 1:
            raise DataError(
                "Requested ``axis`` property for a box monitor; ``axis`` is defined "
                "for surface monitors only."
            )
        return self.size.index(0.0)

    @property
    def local_origin(self) -> Coordinate:
        """Returns the local origin associated with this monitor."""
        if self.custom_origin is None:
            return self.center
        return self.custom_origin


class Near2FarAngleMonitor(AbstractNear2FarMonitor):
    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
    and invokes the computation of far fields at given observation angles.

    Example
    -------
    >>> monitor = Near2FarAngleMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     freqs=[250e12, 300e12],
    ...     name='n2f_monitor',
    ...     custom_origin=(1,2,3),
    ...     phi=[0, np.pi/2],
    ...     theta=np.linspace(-np.pi/2, np.pi/2, 100)
    ...     )
    """

    theta: ObsGridArray = pydantic.Field(
        ...,
        title="Polar Angles",
        description="Polar angles relative to ``local_origin`` at which to compute far fields.",
        units=RADIAN,
    )

    phi: ObsGridArray = pydantic.Field(
        ...,
        title="Azimuth Angles",
        description="Azimuth angles relative to ``local_origin`` at which to compute far fields.",
        units=RADIAN,
    )

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 complex number per pair of angles, per frequency,
        # for N_theta, N_phi, L_theta, and L_phi (4 components)
        return BYTES_COMPLEX * len(self.theta) * len(self.phi) * len(self.freqs) * 4


class Near2FarCartesianMonitor(AbstractNear2FarMonitor):
    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
    and invokes the computation of far fields on a Cartesian observation plane.
    The far-field approximation is used, so results projected to points close to
    the monitor's position, compared to its size, may not be accurate.

    Example
    -------
    >>> monitor = Near2FarCartesianMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     freqs=[250e12, 300e12],
    ...     name='n2f_monitor',
    ...     custom_origin=(1,2,3),
    ...     x=[-1, 0, 1],
    ...     y=[-2, -1, 0, 1, 2],
    ...     plane_axis=2,
    ...     plane_distance=5
    ...     )
    """

    plane_axis: Axis = pydantic.Field(
        ...,
        title="Observation plane axis",
        description="Axis along which the observation plane is oriented.",
    )

    plane_distance: float = pydantic.Field(
        ...,
        title="Observation plane signed distance",
        description="Signed distance of the observation plane along ``plane_axis`` "
        "w.r.t. ``local_origin``",
    )

    x: ObsGridArray = pydantic.Field(
        ...,
        title="Local x observation coordinates",
        description="Local x observation coordinates w.r.t. ``local_origin`` and ``plane_axis``. "
        "When ``plane_axis`` is 0, this corresponds to the global y axis. "
        "When ``plane_axis`` is 1, this corresponds to the global x axis. "
        "When ``plane_axis`` is 2, this corresponds to the global x axis. ",
        units=MICROMETER,
    )

    y: ObsGridArray = pydantic.Field(
        ...,
        title="Local y observation coordinates",
        description="Local y observation coordinates w.r.t. ``local_origin`` and ``plane_axis``. "
        "When ``plane_axis`` is 0, this corresponds to the global z axis. "
        "When ``plane_axis`` is 1, this corresponds to the global z axis. "
        "When ``plane_axis`` is 2, this corresponds to the global y axis. ",
        units=MICROMETER,
    )

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 complex number per pair of grid points, per frequency,
        # for N_theta, N_phi, L_theta, and L_phi (4 components)
        return BYTES_COMPLEX * len(self.x) * len(self.y) * len(self.freqs) * 4


class Near2FarKSpaceMonitor(AbstractNear2FarMonitor):
    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
    and invokes the computation of far fields on an observation plane defined in k-space.

    Example
    -------
    >>> monitor = Near2FarKSpaceMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     freqs=[250e12, 300e12],
    ...     name='n2f_monitor',
    ...     custom_origin=(1,2,3),
    ...     u_axis=2,
    ...     ux=[1,2],
    ...     uy=[3,4,5]
    ...     )
    """

    u_axis: Axis = pydantic.Field(
        ...,
        title="Observation plane axis",
        description="Axis along which the observation plane is oriented.",
    )

    ux: ObsGridArray = pydantic.Field(
        ...,
        title="Normalized kx",
        description="Local x component of wave vectors on the observation plane, "
        "relative to ``local_origin`` and oriented with respect to ``u_axis``, "
        "normalized by (2*pi/lambda) where lambda is the wavelength "
        "associated with the background medium.",
    )

    uy: ObsGridArray = pydantic.Field(
        ...,
        title="Normalized ky",
        description="Local y component of wave vectors on the observation plane, "
        "relative to ``local_origin`` and oriented with respect to ``u_axis``, "
        "normalized by (2*pi/lambda) where lambda is the wavelength "
        "associated with the background medium.",
    )

    def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
        """Size of monitor storage given the number of points after discretization."""
        # stores 1 complex number per pair of grid points, per frequency,
        # for N_theta, N_phi, L_theta, and L_phi (4 components)
        return BYTES_COMPLEX * len(self.ux) * len(self.uy) * len(self.freqs) * 4


# types of monitors that are accepted by simulation
MonitorType = Union[
    FieldMonitor,
    FieldTimeMonitor,
    PermittivityMonitor,
    FluxMonitor,
    FluxTimeMonitor,
    ModeMonitor,
    ModeSolverMonitor,
    Near2FarAngleMonitor,
    Near2FarCartesianMonitor,
    Near2FarKSpaceMonitor,
]



""" Defines classes specifying meshing in 1D and a collective class for 3D """
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Tuple, List, Union

import numpy as np
import pydantic as pd

from .grid import Coords1D, Coords, Grid
from .mesher import GradedMesher, MesherType
from ..base import Tidy3dBaseModel
from ..types import Axis
from ..boundary import Symmetry
from ..source import SourceType
from ..structure import Structure
from ..geometry import Box
from ...log import SetupError, log
from ...constants import C_0, MICROMETER


class GridSpec1d(Tidy3dBaseModel, ABC):

    """Abstract base class, defines 1D grid generation specifications."""

    def make_coords(  # pylint:disable = too-many-arguments
        self,
        axis: Axis,
        structures: List[Structure],
        symmetry: Tuple[Symmetry, Symmetry, Symmetry],
        wavelength: pd.PositiveFloat,
        num_pml_layers: Tuple[pd.NonNegativeInt, pd.NonNegativeInt],
    ) -> Coords1D:
        """Generate 1D coords to be used as grid boundaries, based on simulation parameters.
        Symmetry, and PML layers will be treated here.

        Parameters
        ----------
        axis : Axis
            Axis of this direction.
        structures : List[Structure]
            List of structures present in simulation, the first one being the simulation domain.
        symmetry : Tuple[Symmetry, Symmetry, Symmetry]
            Reflection symmetry across a plane bisecting the simulation domain
            normal to each of the three axes.
        wavelength : float
            Free-space wavelength.
        num_pml_layers : Tuple[int, int]
            number of layers in the absorber + and - direction along one dimension.

        Returns
        -------
        :class:`.Coords1D`:
            1D coords to be used as grid boundaries.
        """

        # Determine if one should apply periodic boundary condition.
        # This should only affect auto nonuniform mesh generation for now.
        is_periodic = sum(num_pml_layers) == 0 and symmetry[axis] == 0

        # generate boundaries
        bound_coords = self._make_coords_initial(
            axis=axis,
            structures=structures,
            wavelength=wavelength,
            symmetry=symmetry,
            is_periodic=is_periodic,
        )

        # incooperate symmetries
        if symmetry[axis] != 0:
            # Offset to center if symmetry present
            center = structures[0].geometry.center[axis]
            center_ind = np.argmin(np.abs(center - bound_coords))
            bound_coords += center - bound_coords[center_ind]
            bound_coords = bound_coords[bound_coords >= center]
            bound_coords = np.append(2 * center - bound_coords[:0:-1], bound_coords)

        # Add PML layers in using dl on edges
        bound_coords = self._add_pml_to_bounds(num_pml_layers, bound_coords)
        return bound_coords

    @abstractmethod
    def _make_coords_initial(
        self,
        axis: Axis,
        structures: List[Structure],
        **kwargs,
    ) -> Coords1D:
        """Generate 1D coords to be used as grid boundaries, based on simulation parameters.
        Symmetry, PML etc. are not considered in this method.

        For auto nonuniform generation, it will take some more arguments.

        Parameters
        ----------
        structures : List[Structure]
            List of structures present in simulation, the first one being the simulation domain.
        **kwargs
            Other arguments

        Returns
        -------
        :class:`.Coords1D`:
            1D coords to be used as grid boundaries.
        """

    @staticmethod
    def _add_pml_to_bounds(num_layers: Tuple[int, int], bounds: Coords1D) -> Coords1D:
        """Append absorber layers to the beginning and end of the simulation bounds
        along one dimension.

        Parameters
        ----------
        num_layers : Tuple[int, int]
            number of layers in the absorber + and - direction along one dimension.
        bound_coords : np.ndarray
            coordinates specifying boundaries between cells along one dimension.

        Returns
        -------
        np.ndarray
            New bound coordinates along dimension taking abosrber into account.
        """
        if bounds.size < 2:
            return bounds

        first_step = bounds[1] - bounds[0]
        last_step = bounds[-1] - bounds[-2]
        add_left = bounds[0] - first_step * np.arange(num_layers[0], 0, -1)
        add_right = bounds[-1] + last_step * np.arange(1, num_layers[1] + 1)
        return np.concatenate((add_left, bounds, add_right))


class UniformGrid(GridSpec1d):

    """Uniform 1D grid.

    Example
    -------
    >>> grid_1d = UniformGrid(dl=0.1)
    """

    dl: pd.PositiveFloat = pd.Field(
        ...,
        title="Grid Size",
        description="Grid size for uniform grid generation.",
        units=MICROMETER,
    )

    def _make_coords_initial(
        self,
        axis: Axis,
        structures: List[Structure],
        **kwargs,
    ) -> Coords1D:
        """Uniform 1D coords to be used as grid boundaries.

        Parameters
        ----------
        axis : Axis
            Axis of this direction.
        structures : List[Structure]
            List of structures present in simulation, the first one being the simulation domain.
        **kwargs:
            Other arguments all go here.

        Returns
        -------
        :class:`.Coords1D`:
            1D coords to be used as grid boundaries.
        """

        center, size = structures[0].geometry.center[axis], structures[0].geometry.size[axis]

        # Take a number of steps commensurate with the size; make dl a bit smaller if needed
        num_cells = int(np.ceil(size / self.dl))

        # Make sure there's at least one cell
        num_cells = max(num_cells, 1)

        # Adjust step size to fit simulation size exactly
        dl_snapped = size / num_cells if size > 0 else self.dl

        return center - size / 2 + np.arange(num_cells + 1) * dl_snapped


class CustomGrid(GridSpec1d):

    """Custom 1D grid supplied as a list of grid cell sizes centered on the simulation center.

    Example
    -------
    >>> grid_1d = CustomGrid(dl=[0.2, 0.2, 0.1, 0.1, 0.1, 0.2, 0.2])
    """

    dl: Tuple[pd.PositiveFloat, ...] = pd.Field(
        ...,
        title="Customized grid sizes.",
        description="An array of custom nonuniform grid sizes. The resulting grid is centered on "
        "the simulation center such that it spans the region "
        "``(center - sum(dl)/2, center + sum(dl)/2)``. "
        "Note: if supplied sizes do not cover the simulation size, the first and last sizes "
        "are repeated to cover the simulation domain.",
        units=MICROMETER,
    )

    def _make_coords_initial(
        self,
        axis: Axis,
        structures: List[Structure],
        **kwargs,
    ) -> Coords1D:
        """Customized 1D coords to be used as grid boundaries.

        Parameters
        ----------
        axis : Axis
            Axis of this direction.
        structures : List[Structure]
            List of structures present in simulation, the first one being the simulation domain.
        *kwargs
            Other arguments all go here.

        Returns
        -------
        :class:`.Coords1D`:
            1D coords to be used as grid boundaries.
        """

        center, size = structures[0].geometry.center[axis], structures[0].geometry.size[axis]

        # get bounding coordinates
        dl = np.array(self.dl)
        bound_coords = np.append(0.0, np.cumsum(dl))

        # place the middle of the bounds at the center of the simulation along dimension
        bound_coords += center - bound_coords[-1] / 2

        # chop off any coords outside of simulation bounds
        bound_min = center - size / 2
        bound_max = center + size / 2
        bound_coords = bound_coords[bound_coords <= bound_max]
        bound_coords = bound_coords[bound_coords >= bound_min]

        # if not extending to simulation bounds, repeat beginning and end
        dl_min = dl[0]
        dl_max = dl[-1]
        while bound_coords[0] - dl_min >= bound_min:
            bound_coords = np.insert(bound_coords, 0, bound_coords[0] - dl_min)
        while bound_coords[-1] + dl_max <= bound_max:
            bound_coords = np.append(bound_coords, bound_coords[-1] + dl_max)

        return bound_coords


class AutoGrid(GridSpec1d):
    """Specification for non-uniform grid along a given dimension.

    Example
    -------
    >>> grid_1d = AutoGrid(min_steps_per_wvl=16, max_scale=1.4)
    """

    min_steps_per_wvl: float = pd.Field(
        10.0,
        title="Minimal number of steps per wavelength",
        description="Minimal number of steps per wavelength in each medium.",
        ge=6.0,
    )

    max_scale: float = pd.Field(
        1.4,
        title="Maximum Grid Size Scaling",
        description="Sets the maximum ratio between any two consecutive grid steps.",
        ge=1.2,
        lt=2.0,
    )

    mesher: MesherType = pd.Field(
        GradedMesher(),
        title="Grid Construction Tool",
        description="The type of mesher to use to generate the grid automatically.",
    )

    def _make_coords_initial(  # pylint:disable=too-many-arguments,arguments-differ,too-many-locals
        self,
        axis: Axis,
        structures: List[Structure],
        wavelength: float,
        symmetry: Symmetry,
        is_periodic: bool,
    ) -> Coords1D:
        """Customized 1D coords to be used as grid boundaries.

        Parameters
        ----------
        axis : Axis
            Axis of this direction.
        structures : List[Structure]
            List of structures present in simulation.
        wavelength : float
            Free-space wavelength.
        symmetry : Tuple[Symmetry, Symmetry, Symmetry]
            Reflection symmetry across a plane bisecting the simulation domain
            normal to each of the three axes.
        is_periodic : bool
            Apply periodic boundary condition or not.

        Returns
        -------
        :class:`.Coords1D`:
            1D coords to be used as grid boundaries.
        """

        sim_cent = list(structures[0].geometry.center)
        sim_size = list(structures[0].geometry.size)
        for dim, sym in enumerate(symmetry):
            if sym != 0:
                sim_cent[dim] += sim_size[dim] / 4
                sim_size[dim] /= 2
        symmetry_domain = Box(center=sim_cent, size=sim_size)

        # New list of structures with symmetry applied
        struct_list = [Structure(geometry=symmetry_domain, medium=structures[0].medium)]
        for structure in structures[1:]:
            if symmetry_domain.intersects(structure.geometry):
                struct_list.append(structure)

        # parse structures
        interval_coords, max_dl_list = self.mesher.parse_structures(
            axis, struct_list, wavelength, self.min_steps_per_wvl
        )
        # Put just a single pixel if 2D-like simulation
        if interval_coords.size == 1:
            dl = wavelength / self.min_steps_per_wvl
            return np.array([sim_cent[axis] - dl / 2, sim_cent[axis] + dl / 2])

        # generate mesh steps
        interval_coords = np.array(interval_coords).flatten()
        max_dl_list = np.array(max_dl_list).flatten()
        len_interval_list = interval_coords[1:] - interval_coords[:-1]
        dl_list = self.mesher.make_grid_multiple_intervals(
            max_dl_list, len_interval_list, self.max_scale, is_periodic
        )

        # generate boundaries
        bound_coords = np.append(0.0, np.cumsum(np.concatenate(dl_list)))
        bound_coords += interval_coords[0]

        # fix simulation domain boundaries which may be slightly off
        domain_bounds = [bound[axis] for bound in symmetry_domain.bounds]
        if not np.all(np.isclose(bound_coords[[0, -1]], domain_bounds)):
            raise SetupError(
                f"AutoGrid coordinates along axis {axis} do not match the simulation "
                "domain, indicating an unexpected error in the meshing. Please create a github "
                "issue so that the problem can be investigated. In the meantime, switch to "
                f"uniform or custom grid along {axis}."
            )
        bound_coords[[0, -1]] = domain_bounds

        return np.array(bound_coords)


GridType = Union[UniformGrid, CustomGrid, AutoGrid]


class GridSpec(Tidy3dBaseModel):

    """Collective grid specification for all three dimensions.

    Example
    -------
    >>> uniform = UniformGrid(dl=0.1)
    >>> custom = CustomGrid(dl=[0.2, 0.2, 0.1, 0.1, 0.1, 0.2, 0.2])
    >>> auto = AutoGrid(min_steps_per_wvl=12)
    >>> grid_spec = GridSpec(grid_x=uniform, grid_y=custom, grid_z=auto, wavelength=1.5)
    """

    grid_x: GridType = pd.Field(
        AutoGrid(),
        title="Grid specification along x-axis",
        description="Grid specification along x-axis",
    )

    grid_y: GridType = pd.Field(
        AutoGrid(),
        title="Grid specification along y-axis",
        description="Grid specification along y-axis",
    )

    grid_z: GridType = pd.Field(
        AutoGrid(),
        title="Grid specification along z-axis",
        description="Grid specification along z-axis",
    )

    wavelength: float = pd.Field(
        None,
        title="Free-space wavelength",
        description="Free-space wavelength for automatic nonuniform grid. It can be 'None' "
        "if there is at least one source in the simulation, in which case it is defined by "
        "the source central frequency.",
        units=MICROMETER,
    )

    override_structures: Tuple[Structure, ...] = pd.Field(
        (),
        title="Grid specification override structures",
        description="A set of structures that is added on top of the simulation structures in "
        "the process of generating the grid. This can be used to refine the grid or make it "
        "coarser depending than the expected need for higher/lower resolution regions.",
    )

    @property
    def auto_grid_used(self) -> bool:
        """True if any of the three dimensions uses :class:`.AutoGrid`."""
        grid_list = [self.grid_x, self.grid_y, self.grid_z]
        return np.any([isinstance(mesh, AutoGrid) for mesh in grid_list])

    def wavelength_from_sources(self, sources: List[SourceType]) -> pd.PositiveFloat:
        """Define a wavelength based on supplied sources. Called if auto mesh is used and
        ``self.wavelength is None``."""

        # Use central frequency of sources, if any.
        freqs = np.array([source.source_time.freq0 for source in sources])
        # no sources
        if len(freqs) == 0:
            raise SetupError(
                "Automatic grid generation requires the input of 'wavelength' or sources."
            )
        # multiple sources of different central frequencies
        if len(freqs) > 0 and not np.all(np.isclose(freqs, freqs[0])):
            raise SetupError(
                "Sources of different central frequencies are supplied. "
                "Please supply a 'wavelength' value for 'grid_spec'."
            )

        return C_0 / freqs[0]

    def make_grid(
        self,
        structures: List[Structure],
        symmetry: Tuple[Symmetry, Symmetry, Symmetry],
        sources: List[SourceType],
        num_pml_layers: List[Tuple[pd.NonNegativeInt, pd.NonNegativeInt]],
    ) -> Grid:
        """Make the entire simulation grid based on some simulation parameters.

        Parameters
        ----------
        structures : List[Structure]
            List of structures present in the simulation. The first structure must be the
            simulation geometry with the simulation background medium.
        symmetry : Tuple[Symmetry, Symmetry, Symmetry]
            Reflection symmetry across a plane bisecting the simulation domain
            normal to each of the three axes.
        sources : List[SourceType]
            List of sources.
        num_pml_layers : List[Tuple[float, float]]
            List containing the number of absorber layers in - and + boundaries.

        Returns
        -------
        Grid:
            Entire simulation grid.
        """

        # Set up wavelength for automatic mesh generation if needed.
        wavelength = self.wavelength
        if self.wavelength is None and self.auto_grid_used:
            wavelength = self.wavelength_from_sources(sources)
            log.info(f"Auto meshing using wavelength {wavelength:1.4f} defined from sources.")

        coords_x = self.grid_x.make_coords(
            axis=0,
            structures=list(structures) + list(self.override_structures),
            symmetry=symmetry,
            wavelength=wavelength,
            num_pml_layers=num_pml_layers[0],
        )
        coords_y = self.grid_y.make_coords(
            axis=1,
            structures=list(structures) + list(self.override_structures),
            symmetry=symmetry,
            wavelength=wavelength,
            num_pml_layers=num_pml_layers[1],
        )
        coords_z = self.grid_z.make_coords(
            axis=2,
            structures=list(structures) + list(self.override_structures),
            symmetry=symmetry,
            wavelength=wavelength,
            num_pml_layers=num_pml_layers[2],
        )

        coords = Coords(x=coords_x, y=coords_y, z=coords_z)
        return Grid(boundaries=coords)

    @classmethod
    def auto(  # pylint:disable=too-many-arguments
        cls,
        wavelength: pd.PositiveFloat = None,
        min_steps_per_wvl: pd.PositiveFloat = 10.0,
        max_scale: pd.PositiveFloat = 1.4,
        override_structures: List[Structure] = (),
        mesher: MesherType = GradedMesher(),
    ) -> GridSpec:
        """Use the same :class:`AutoGrid` along each of the three directions.

        Parameters
        ----------
        wavelength : pd.PositiveFloat, optional
            Free-space wavelength for automatic nonuniform grid. It can be 'None'
            if there is at least one source in the simulation, in which case it is defined by
            the source central frequency.
        min_steps_per_wvl : pd.PositiveFloat, optional
            Minimal number of steps per wavelength in each medium.
        max_scale : pd.PositiveFloat, optional
            Sets the maximum ratio between any two consecutive grid steps.
        override_structures : List[Structure]
            A list of structures that is added on top of the simulation structures in
            the process of generating the grid. This can be used to refine the grid or make it
            coarser depending than the expected need for higher/lower resolution regions.
        mesher : MesherType = GradedMesher()
            The type of mesher to use to generate the grid automatically.

        Returns
        -------
        GridSpec
            :class:`GridSpec` with the same automatic nonuniform grid settings in each direction.
        """

        grid_1d = AutoGrid(min_steps_per_wvl=min_steps_per_wvl, max_scale=max_scale, mesher=mesher)
        return cls(
            wavelength=wavelength,
            grid_x=grid_1d,
            grid_y=grid_1d,
            grid_z=grid_1d,
            override_structures=override_structures,
        )

    @classmethod
    def uniform(cls, dl: float) -> GridSpec:
        """Use the same :class:`UniformGrid` along each of the three directions.

        Parameters
        ----------
        dl : float
            Grid size for uniform grid generation.

        Returns
        -------
        GridSpec
            :class:`GridSpec` with the same uniform grid size in each direction.
        """

        grid_1d = UniformGrid(dl=dl)
        return cls(grid_x=grid_1d, grid_y=grid_1d, grid_z=grid_1d)



"""Defines how the specific data objects render as UI components."""
from abc import ABC
from typing import Union, Tuple, List, Dict

import numpy as np
import plotly.graph_objects as go
import pydantic as pd
from dash import dcc, html, Input
from typing_extensions import Literal

from .component import UIComponent
from .utils import PlotlyFig
from ...components.data import FluxData, FluxTimeData, FieldData, FieldTimeData
from ...components.data import ModeSolverData, ModeData, ScalarFieldDataArray
from ...components.geometry import Geometry
from ...components.types import Axis, Direction
from ...log import Tidy3dKeyError, log

# divide an SI unit by these to put in the unit in the name
#    eg.  1e-12 (sec) / PICOSECOND = 1 (ps)
# multiply a value in the specified unit by these to put in SI
#    eg.  1 (THz) * TERAHERTZ = 1e12 (Hz)
PICOSECOND = 1e-12
TERAHERTZ = 1e12

# supported data types
Tidy3dBaseModelType = Union[
    FluxData, FluxTimeData, FieldData, FieldTimeData, ModeData, ModeSolverData
]


class DataPlotly(UIComponent, ABC):
    """Base class for anything that generates dash components from tidy3d data objects."""

    monitor_name: str = pd.Field(..., tite="monitor name", description="Name of the monitor.")
    data: Tidy3dBaseModelType = pd.Field(
        ..., tite="data", description="The Tidy3dBaseModel object wrapped by this UI component."
    )

    @property
    def label(self) -> str:
        """Get tab label for this component."""
        return f"monitor: '{self.monitor_name}'"

    # @property
    # def unique_id(self) -> str:
    #     """Get unique id for this component."""
    #     return f"monitor_{self.monitor_name}"

    @staticmethod
    def sel_by_val(data, val: str) -> Tidy3dBaseModelType:
        """Select the correct data type based on the `val` selection."""
        if val.lower() == "real":
            return data.real
        if val.lower() == "imag":
            return data.imag
        if val.lower() == "abs":
            return abs(data)
        if val.lower() == "abs^2":
            return abs(data) ** 2
        raise ValueError(f"Could not find the right function to apply with {val}.")

    def append_monitor_name(self, value: str) -> Dict[str, str]:
        """Adds the monitor name to a value, used to make the ids unique across all monitors."""
        return {"type": f"{type(self.data).__name__}_{value}", "name": self.monitor_name}

    @classmethod
    def from_monitor_data(
        cls, monitor_name: str, monitor_data: Tidy3dBaseModelType, **kwargs
    ) -> Tidy3dBaseModelType:
        """Load a PlotlyData UI component from the monitor name and its data."""

        # maps the supplied ``monitor_data`` argument to the corresponding plotly wrapper.
        data_plotly_map = (
            (FluxData, FluxDataPlotly),
            (FluxTimeData, FluxTimeDataPlotly),
            (FieldData, FieldDataPlotly),
            (FieldTimeData, FieldTimeDataPlotly),
            (ModeSolverData, ModeSolverDataPlotly),
            (ModeData, ModeDataPlotly),
        )

        # get the type of the supplied ``monitor_data``.
        monitor_data_type = type(monitor_data)

        for mnt_data, mnt_data_plotly in data_plotly_map:
            if isinstance(monitor_data, mnt_data):
                return mnt_data_plotly(data=monitor_data, monitor_name=monitor_name, **kwargs)

        log.warning(
            f"could not find a plotly wrapper for monitor {monitor_name}"
            f"of type {monitor_data_type.__name__}"
        )
        return None


class AbstractFluxDataPlotly(DataPlotly, ABC):
    """Flux data in frequency or time domain."""

    data: Union[FluxData, FluxTimeData] = pd.Field(
        ...,
        title="data",
        description="A flux data object in freq or time domain.",
    )

    @property
    def ft_label_coords_units(self) -> Tuple[str, List[float], str]:
        """Get the ``freq`` or ``time`` label, coords, and units depending on the data contents."""

        if "f" in self.data.data.coords:
            ft_label = "freq"
            ft_coords = self.data.data.coords["f"].values / TERAHERTZ
            ft_units = "THz"
        elif "t" in self.data.data.coords:
            ft_label = "time"
            ft_coords = self.data.data.coords["t"].values / PICOSECOND
            ft_units = "ps"
        else:
            raise Tidy3dKeyError("Neither frequency nor time data found in this data object.")

        return ft_label, ft_coords, ft_units

    def make_figure(self) -> PlotlyFig:
        """Generate plotly figure from the current state of self."""
        return self.plotly()

    def make_component(self) -> dcc.Tab:
        """Creates the dash component for this montor data."""

        # initital setup
        fig = self.make_figure()

        # individual components
        flux_plot = html.Div(
            [
                dcc.Graph(
                    id=self.append_monitor_name("figure"),
                    figure=fig,
                )
            ],
            style={"padding": 10, "flex": 1},
        )

        # return the layout of the component so the app can insert it
        return dcc.Tab(
            [
                html.H1(f"Viewing data for {type(self.data).__name__}: '{self.monitor_name}'"),
                flux_plot,
            ],
            label=self.label,
        )

    def plotly(self) -> PlotlyFig:
        """Generate the plotly figure for this component."""

        ft_label, ft_coords, ft_units = self.ft_label_coords_units
        ft_units = "THz" if "f" in ft_label else "ps"

        fig = go.Figure(go.Scatter(x=ft_coords, y=self.data.data.values))

        fig.update_layout(
            xaxis_title=f"{ft_label} ({ft_units})",
            yaxis_title="Flux (normalized)",
        )

        fig.update_layout(yaxis=dict(showexponent="all", exponentformat="e"))

        return fig


class FluxDataPlotly(AbstractFluxDataPlotly):
    """Flux in frequency domain."""

    data: FluxData = pd.Field(
        ...,
        title="data",
        description="A flux data object in the frequency domain.",
    )


class FluxTimeDataPlotly(AbstractFluxDataPlotly):
    """Flux in time domain."""

    data: FluxTimeData = pd.Field(
        ...,
        title="data",
        description="A flux data object in the time domain.",
    )


class ModeDataPlotly(DataPlotly):
    """Flux data in frequency or time domain."""

    data: ModeData = pd.Field(
        ...,
        title="data",
        description="A mode amplitude data object",
    )

    amps_or_neff: Literal["amps", "neff"] = pd.Field(
        "amps",
        title="Amps or effective index value",
        description="The state of the component's 'amplitude or neff' value.",
    )

    val: str = pd.Field(
        "abs^2",
        title="Plotting value value",
        description="The state of the component's plotting value value.",
    )

    dir_val: Direction = pd.Field(
        None, title="Direction value", description="The state of the component's direction value."
    )

    mode_ind_val: int = pd.Field(
        None, title="Mode index value", description="The state of the component's mode index value."
    )

    @property
    def mode_ind_coords(self) -> List[float]:
        """Get the mode indices."""
        return self.data.amps.coords["mode_index"].values

    @property
    def direction_coords(self) -> List[float]:
        """Get the mode indices."""
        return self.data.amps.coords["direction"].values

    @property
    def ft_label_coords_units(self) -> Tuple[str, List[float], str]:
        """Get the `freq` or `time` label and coords."""

        ft_label = "freq"
        ft_coords = self.data.amps.coords["f"].values / TERAHERTZ
        ft_units = "THz"
        return ft_label, ft_coords, ft_units

    @property
    def dir_dropdown_hidden(self) -> bool:
        """Should the dropdown be hidden?"""
        return self.amps_or_neff == "neff"

    def make_figure(self) -> PlotlyFig:
        """Generate plotly figure from the current state of self."""

        if self.dir_val is None:
            self.dir_val = self.direction_coords[0]

        if self.mode_ind_val is None:
            self.mode_ind_val = self.mode_ind_coords[0]

        if self.amps_or_neff == "amps":
            return self.plotly_amps(mode_index=self.mode_ind_val, dir_val=self.dir_val)

        return self.plotly_neff(mode_index=self.mode_ind_val)

    def make_component(self) -> dcc.Tab:
        """Creates the dash component for this montor data."""

        # initital setup
        fig = self.make_figure()

        # individual components

        # amplitude plot
        amp_plot = html.Div(
            [
                dcc.Graph(
                    id=self.append_monitor_name("figure"),
                    figure=fig,
                )
            ],
            style={"padding": 10, "flex": 1},
        )

        # select amps or neff
        amps_or_neff_dropdown = dcc.Dropdown(
            options=["amps", "neff"],
            value=self.amps_or_neff,
            id=self.append_monitor_name("amps_or_neff_dropdown"),
        )

        # select real, abs, imag, power
        field_value_dropdown = dcc.Dropdown(
            options=["real", "imag", "abs^2"],
            value=self.val,
            id=self.append_monitor_name("val_dropdown"),
        )

        # header for direction dropdown
        dir_dropdown_header = html.Div(
            [html.H2("Direction: forward (+) or backward (-).")],
            hidden=self.dir_dropdown_hidden,
            id=self.append_monitor_name("dir_dropdown_header"),
        )

        # select direction
        dir_value_dropdown = html.Div(
            [
                dcc.Dropdown(
                    options=list(self.direction_coords),
                    value=self.dir_val,
                    id=self.append_monitor_name("dir_dropdown"),
                )
            ],
            hidden=self.dir_dropdown_hidden,
            id=self.append_monitor_name("dir_dropdown_div"),
        )

        # mode index selector
        mode_ind_dropdown = html.Div(
            [
                dcc.Dropdown(
                    options=list(self.mode_ind_coords),
                    value=self.mode_ind_val,
                    id=self.append_monitor_name("mode_index_selector"),
                ),
            ]
        )

        # data control panel
        panel_children = [
            html.H2("Amplitude or effective index."),
            amps_or_neff_dropdown,
            html.H2("Value to plot."),
            field_value_dropdown,
            dir_dropdown_header,
            dir_value_dropdown,
            html.H2("Mode Index."),
            mode_ind_dropdown,
        ]

        plot_selections = html.Div(panel_children, style={"padding": 10, "flex": 1})

        # define layout
        component = dcc.Tab(
            [
                html.H1(f"Viewing data for {type(self.data).__name__}: '{self.monitor_name}'"),
                html.Div(
                    [
                        # left hand side
                        amp_plot,
                        # right hand side
                        plot_selections,
                    ],
                    # make elements in above list stack row-wise
                    style={"display": "flex", "flexDirection": "row"},
                ),
            ],
            label=self.label,
        )

        return component

    def plotly_amps(self, mode_index: int, dir_val: str):
        """Make a line chart for the mode amplitudes."""

        ft_label, ft_coords, ft_units = self.ft_label_coords_units
        ft_units = "THz" if "f" in ft_label else "ps"

        amp_val = self.sel_by_val(self.data.amps, val=self.val)
        amp_val = amp_val.sel(direction=dir_val, mode_index=mode_index)
        fig = go.Figure(go.Scatter(x=ft_coords, y=amp_val))

        fig.update_layout(
            title_text=f"amplitudes of mode w/ index {self.mode_ind_val} in {dir_val} direction.",
            title_x=0.5,
            xaxis_title=f"{ft_label} ({ft_units})",
            yaxis_title="Amplitude (normalized)",
        )
        return fig

    def plotly_neff(self, mode_index: int):
        """Make a line chart for the mode amplitudes."""

        ft_label, ft_coords, ft_units = self.ft_label_coords_units
        ft_units = "THz" if "f" in ft_label else "ps"

        neff_val = self.sel_by_val(self.data.n_complex, val=self.val)
        neff_val = neff_val.sel(mode_index=mode_index)
        fig = go.Figure(go.Scatter(x=ft_coords, y=neff_val))

        fig.update_layout(
            title_text=f"effective index of mode w/ index {self.mode_ind_val}",
            title_x=0.5,
            xaxis_title=f"{ft_label} ({ft_units})",
            yaxis_title="Effective index",
        )
        return fig


class AbstractFieldDataPlotly(DataPlotly, ABC):
    """Some kind of field-like data plotted in the app."""

    data: Union[FieldData, FieldTimeData] = pd.Field(
        ...,
        title="data",
        description="A Field-like data object.",
    )

    field_val: str = pd.Field(
        None, title="Field value", description="The component's field component value."
    )

    val: str = pd.Field(
        "abs", title="Plot value value", description="The component's plotting value value."
    )

    cs_axis: Axis = pd.Field(
        None,
        title="Cross section axis value",
        description="The component's cross section axis value.",
    )

    cs_val: float = pd.Field(
        None,
        title="Cross section position value",
        description="The component's cross section position value.",
    )

    ft_val: float = pd.Field(
        None, title="Freq or time value", description="The component's frequency or time value."
    )

    mode_ind_val: int = pd.Field(
        None, title="Mode index value", description="The component's mode index value value."
    )

    @property
    def ft_label_coords_units(self) -> Tuple[str, List[float], str]:
        """Get the `freq` or `time` label and coords."""

        scalar_field_data = self.data.data_dict[self.field_val]

        if "f" in self.scalar_field_data.data.coords:
            ft_label = "freq"
            ft_coords = scalar_field_data.data.coords["f"].values / TERAHERTZ
            ft_units = "THz"
        elif "t" in self.scalar_field_data.data.coords:
            ft_label = "time"
            ft_coords = scalar_field_data.data.coords["t"].values / PICOSECOND
            ft_units = "ps"
        else:
            raise Tidy3dKeyError("neither frequency nor time data found in this data object.")

        return ft_label, ft_coords, ft_units

    @property
    def inital_field_val(self):
        """The starting field value."""
        field_vals = list(self.data.data_dict.keys())
        if not field_vals:
            raise ValueError("Data doesn't have any field components stored.")
        return field_vals[0]

    @property
    def scalar_field_data(self) -> ScalarFieldDataArray:
        """The current scalar field monitor data."""
        if self.field_val is None:
            self.field_val = self.inital_field_val

        return self.data.data_dict[self.field_val]

    @property
    def inital_cs_axis(self) -> List[int]:
        """Returns the cross section axis that plots the 2D view."""
        coords = self.scalar_field_data.data.coords
        coords_xyz = [coords[xyz_label].values for xyz_label in "xyz"]
        has_volume = [len(coord) > 3 for coord in coords_xyz]

        # if a 2D view
        if sum(has_volume) == 2:
            # initialize with the cross section axis set up to display the 2D plot
            return has_volume.index(False)

        # otherwise, just initialize with x as cross section axis.
        return 0

    @property
    def xyz_label_coords(self) -> Tuple[str, List[float]]:
        """Get the plane normal direction label and coords."""
        if self.cs_axis is None:
            self.cs_axis = self.inital_cs_axis
        xyz_label = "xyz"[self.cs_axis]
        xyz_coords = self.scalar_field_data.data.coords[xyz_label].values
        return xyz_label, xyz_coords

    @property
    def mode_ind_coords(self) -> List[int]:
        """Get the mode indices."""
        if "mode_index" in self.scalar_field_data.data.coords:
            return self.scalar_field_data.data.coords["mode_index"].values
        return None

    def make_figure(self) -> PlotlyFig:
        """Generate plotly figure from the current state of self."""

        # if no field specified, use the first one in the fields list
        if self.field_val is None:
            self.field_val = self.inital_field_val

        if self.cs_axis is None:
            self.cs_axis = self.inital_cs_axis

        # if no mode_ind_val specified, use the first of the coords (or None)
        if self.mode_ind_val is None:
            mode_indices = self.mode_ind_coords
            self.mode_ind_val = mode_indices[0] if mode_indices is not None else None

        # if cross section value, use the average of the coordinates of the current axis
        xyz_label, xyz_coords = self.xyz_label_coords
        if self.cs_val is None:
            self.cs_val = np.mean(xyz_coords)

        # if no freq or time value, use the average of the coordinates
        ft_label, ft_coords, _ = self.ft_label_coords_units
        if self.ft_val is None:
            self.ft_val = np.mean(ft_coords)

        plotly_kwargs = {
            xyz_label: self.cs_val,
            "field": self.field_val,
            ft_label: self.ft_val,
            "val": self.val,
            "mode_index": self.mode_ind_val,
        }

        return self.plotly(**plotly_kwargs)

    def make_component(self) -> dcc.Tab:  # pylint:disable=too-many-locals
        """Creates the dash component."""

        # initial setup
        xyz_label, xyz_coords = self.xyz_label_coords
        ft_label, ft_coords, ft_units = self.ft_label_coords_units
        fig = self.make_figure()

        # individual components

        # plot of the fields
        field_plot = html.Div(
            [
                dcc.Graph(
                    id=self.append_monitor_name("figure"),
                    figure=fig,
                )
            ],
            style={"padding": 10, "flex": 1},
        )

        # pick the field component
        field_dropdown = dcc.Dropdown(
            options=list(self.data.data_dict.keys()),
            value=self.field_val,
            id=self.append_monitor_name("field_dropdown"),
        )

        # pick the real / imag / abs to plot
        field_value_dropdown = dcc.Dropdown(
            options=["real", "imag", "abs"],
            value=self.val,
            id=self.append_monitor_name("val_dropdown"),
        )

        # pick the cross section axis
        xyz_dropdown = dcc.Dropdown(
            options=["x", "y", "z"],
            value=xyz_label,
            id=self.append_monitor_name("cs_axis_dropdown"),
        )

        # pick the cross section position
        xyz_slider = dcc.Slider(
            min=xyz_coords[0],
            max=xyz_coords[-1],
            value=self.cs_val,
            id=self.append_monitor_name("cs_slider"),
        )

        # combine the cross section selection into one component
        xyz_selection = html.Div([xyz_dropdown, xyz_slider])

        # pick the frequency or time
        freq_time_slider = html.Div(
            [
                dcc.Slider(
                    min=ft_coords[0],
                    max=ft_coords[-1],
                    value=self.ft_val,
                    id=self.append_monitor_name("ft_slider"),
                ),
            ]
        )

        # all the controls for adjusting plotted data
        plot_selections = html.Div(
            [
                html.H2("Field component."),
                field_dropdown,
                html.H2("Value to plot."),
                field_value_dropdown,
                html.H2("Cross-section axis and position."),
                xyz_selection,
                html.H2(f"{ft_label} value ({ft_units})."),
                freq_time_slider,
            ],
            style={"padding": 10, "flex": 1},
        )

        # add a mode index dropdown to right hand side, if applicable
        if self.mode_ind_val is not None:
            # make a mode index label and dropdown
            mode_ind_label = html.H2("Mode Index component.")
            mode_ind_dropdown = html.Div(
                [
                    dcc.Dropdown(
                        options=list(self.mode_ind_coords),
                        value=self.mode_ind_val,
                        id=self.append_monitor_name("mode_index_selector"),
                    ),
                ]
            )

            # add these to the plot selections panel component
            plot_selections.children.append(mode_ind_label)
            plot_selections.children.append(mode_ind_dropdown)

        # full layout
        component = dcc.Tab(
            [
                # title
                html.H1(f"Viewing data for {type(self.data).__name__}: '{self.monitor_name}'"),
                # below title
                html.Div(
                    [
                        # left hand side
                        field_plot,
                        # right hand side
                        plot_selections,
                    ],
                    # make elements in above list stack row-wise
                    style={"display": "flex", "flexDirection": "row"},
                ),
            ],
            # label for the tab
            label=self.label,
        )

        # these are the inputs to the callback function which links the buttons to the figure
        app_inputs = [
            Input(self.append_monitor_name("field_dropdown"), "value"),
            Input(self.append_monitor_name("val_dropdown"), "value"),
            Input(self.append_monitor_name("cs_axis_dropdown"), "value"),
            Input(self.append_monitor_name("cs_slider"), "value"),
            Input(self.append_monitor_name("ft_slider"), "value"),
        ]

        # add the mode index dropdown to the app inputs, if defined
        if self.mode_ind_val is not None:
            app_inputs.append(Input(self.append_monitor_name("mode_index_selector"), "value"))

        # link what happens in the app_inputs to what gets displayed in the figure

        return component

    def plotly(  # pylint:disable=too-many-arguments, too-many-locals
        self,
        field: str,
        val: Literal["real", "imag", "abs"],
        freq: float = None,
        time: float = None,
        x: float = None,
        y: float = None,
        z: float = None,
        mode_index: int = None,
    ) -> PlotlyFig:
        """Creates the plotly figure given some parameters."""

        axis, position = Geometry.parse_xyz_kwargs(x=x, y=y, z=z)

        # grab by field name
        scalar_field_data = self.scalar_field_data.data

        # select mode_index, if given
        if mode_index is not None and "mode_index" in scalar_field_data.coords:
            scalar_field_data = scalar_field_data.sel(mode_index=mode_index)

        # select by frequency, if given
        if freq is not None:
            freq *= TERAHERTZ
            sel_ft = scalar_field_data.sel(f=freq, method="nearest")

        # select by time, if given
        if time is not None:
            time *= PICOSECOND
            sel_ft = scalar_field_data.sel(t=time, method="nearest")

        # select the cross sectional plane data
        xyz_labels = ["x", "y", "z"]
        xyz_kwargs = {xyz_labels.pop(axis): position}
        sel_xyz = sel_ft.sel(**xyz_kwargs, method="nearest")

        # get the correct field value (real, imaginary, abs)
        sel_val = self.sel_by_val(data=sel_xyz, val=val)

        # get the correct x and y labels
        coords_plot_x = sel_val.coords[xyz_labels[0]]
        coords_plot_y = sel_val.coords[xyz_labels[1]]

        # construct the field plot
        fig = go.Figure(
            data=go.Heatmap(
                x=coords_plot_x,
                y=coords_plot_y,
                z=sel_val.values,
                transpose=True,
                type="heatmap",
                colorscale="magma" if val in "abs" in val else "RdBu",
            )
        )

        # update title and x and y labels.
        ft_text = f"f={freq:.2e}" if freq is not None else f"t={time:.2e}"
        _, (xlabel, ylabel) = Geometry.pop_axis("xyz", axis=axis)
        fig.update_layout(
            title_text=f'{val}[{field}({"xyz"[axis]}={position:.2e}, {ft_text})]',
            title_x=0.5,
            xaxis_title=f"{xlabel} (um)",
            yaxis_title=f"{ylabel} (um)",
        )

        return fig


class FieldDataPlotly(AbstractFieldDataPlotly):
    """Plot :class:`.FieldData` in app."""

    data: FieldData = pd.Field(
        ...,
        title="data",
        description="A field data object in the frequency domain",
    )


class FieldTimeDataPlotly(AbstractFieldDataPlotly):
    """Plot :class:`.FieldTimeData` in app."""

    data: FieldTimeData = pd.Field(
        ...,
        title="data",
        description="A field data object in the time domain.",
    )


class ModeSolverDataPlotly(AbstractFieldDataPlotly):
    """Plot :class:`.ModeSolverData` in app."""

    data: ModeSolverData = pd.Field(
        ...,
        title="data",
        description="A mode field object.",
    )



"""Simulation and Geometry plotting with plotly."""
# pylint:disable=too-many-arguments, protected-access
from typing import Tuple

import numpy as np
import plotly.graph_objects as go
from dash import dcc, html
from shapely.geometry.base import BaseGeometry as ShapelyGeo
import pydantic as pd

from .component import UIComponent
from .utils import PlotlyFig, add_fig_if_none, equal_aspect_plotly, plot_params_sim_boundary
from ...components.base import Tidy3dBaseModel
from ...components.geometry import Geometry, Box
from ...components.medium import Medium
from ...components.simulation import Simulation
from ...components.structure import Structure
from ...components.types import Axis
from ...components.viz import PlotParams, plot_params_pml


def plotly_shape(
    shape: ShapelyGeo,
    plot_params: PlotParams,
    fig: PlotlyFig,
    name: str = None,
) -> PlotlyFig:
    """Plot a shape to a figure."""
    _shape = Geometry.evaluate_inf_shape(shape)
    exterior_coords, _ = Geometry.strip_coords(shape=_shape)
    xs, ys = list(zip(*exterior_coords))
    plotly_trace = go.Scatter(
        x=xs,
        y=ys,
        fill="toself",
        fillcolor=plot_params.facecolor,
        line=dict(width=plot_params.linewidth, color=plot_params.facecolor),
        marker=dict(size=0.0001, line=dict(width=0)),
        name=name,
        opacity=plot_params.alpha,
    )
    fig.add_trace(plotly_trace)
    return fig


class GeometryPlotly(Tidy3dBaseModel):
    """Geometry that adds plotly-based implementations of its standard plotting functions."""

    geometry: Geometry

    @equal_aspect_plotly
    @add_fig_if_none
    def plotly(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
        name: str = None,
        **patch_kwargs,
    ) -> PlotlyFig:
        """Plot cross sections on plane using plotly."""

        plot_params = self.geometry.plot_params.copy(deep=True)
        plot_params_struct = plot_params.include_kwargs(**patch_kwargs)

        # for each intersection, plot the shape
        for shape in self.geometry.intersections(x=x, y=y, z=z):
            fig = plotly_shape(
                shape=shape,
                plot_params=plot_params_struct,
                fig=fig,
                name=name,
            )

        return fig


class StructurePlotly(Tidy3dBaseModel):
    """Structure that adds plotly-based implementations of its standard plotting functions."""

    structure: Structure

    @add_fig_if_none
    def plotly(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
        name: str = None,
        **patch_kwargs,
    ) -> PlotlyFig:
        """Use plotly to plot structure's geometric cross section at single (x,y,z) coordinate."""
        geometry = GeometryPlotly(geometry=self.structure.geometry)
        return geometry.plotly(x=x, y=y, z=z, fig=fig, name=name, **patch_kwargs)


class SimulationPlotly(UIComponent):
    """Simulation that adds plotly-based implementations of its standard plotting functions."""

    simulation: Simulation = pd.Field(
        ..., title="Simulation", description="The Simulation instance to plot."
    )

    cs_axis: Axis = pd.Field(
        0,
        title="Cross-section axis",
        description="The axis (0,1,2) representing the plotting plane normal direction.",
    )

    cs_val: float = pd.Field(
        None,
        title="Cross-section value",
        description="The position along the plotting plane axis normal.",
    )

    @property
    def xyz_label_bounds(self):
        """Get the plot normal axis label and the min and max bounds."""

        xyz_label = "xyz"[self.cs_axis]
        bmin, bmax = self.simulation.bounds
        xyz_min = bmin[self.cs_axis]
        xyz_max = bmax[self.cs_axis]
        return xyz_label, (xyz_min, xyz_max)

    def make_figure(self):
        """Generate plotly figure from the current state of self."""

        xyz_label, (xyz_min, xyz_max) = self.xyz_label_bounds
        if self.cs_val is None:
            self.cs_val = (xyz_min + xyz_max) / 2.0
        plotly_kwargs = {xyz_label: self.cs_val}
        return self.plotly(**plotly_kwargs)

    def make_component(self):  # pylint: disable=too-many-locals
        """Creates the dash component."""

        xyz_label, (xyz_min, xyz_max) = self.xyz_label_bounds
        figure = self.make_figure()

        graph = html.Div(
            [dcc.Graph(figure=figure, id="simulation_plot")], style={"padding": 10, "flex": 1}
        )

        xyz_header = html.H2("Cross-section axis and position.")

        xyz_dropdown = dcc.Dropdown(
            options=["x", "y", "z"],
            value=xyz_label,
            id="simulation_cs_axis_dropdown",
        )

        xyz_slider = dcc.Slider(
            min=xyz_min,
            max=xyz_max,
            value=self.cs_val,
            id="simulation_cs_slider",
        )

        xyz_selection = html.Div(
            [xyz_header, xyz_dropdown, xyz_slider],
            style={"padding": 50, "flex": 1},
        )

        return dcc.Tab(
            [
                html.H1("Viewing Simulation."),
                html.Div([graph, xyz_selection], style={"display": "flex", "flexDirection": "row"}),
            ],
            label="Simulation",
        )

    @equal_aspect_plotly
    @add_fig_if_none
    def plotly(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.
        Uses plotly.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.

        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """

        # fig = self._plotly_bounding_box(x=x, y=y, z=z, fig=fig)
        fig = self.plotly_structures(x=x, y=y, z=z, fig=fig)
        fig = self.plotly_sources(x=x, y=y, z=z, fig=fig)
        fig = self.plotly_monitors(x=x, y=y, z=z, fig=fig)
        fig = self.plotly_symmetries(x=x, y=y, z=z, fig=fig)
        fig = self.plotly_pml(x=x, y=y, z=z, fig=fig)
        fig = self._plotly_cleanup(x=x, y=y, z=z, fig=fig)

        return fig

    @equal_aspect_plotly
    @add_fig_if_none
    def plotly_structures(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's structures on a plane defined by one nonzero x,y,z .

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.

        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """

        structures = self.simulation.structures
        medium_shapes = self.simulation._get_structures_plane(structures=structures, x=x, y=y, z=z)
        for (medium, shape) in medium_shapes:
            fig = self._plotly_shape_structure(medium=medium, shape=shape, fig=fig)
        return fig

    def _plotly_shape_structure(
        self, medium: Medium, shape: ShapelyGeo, fig: PlotlyFig
    ) -> PlotlyFig:
        """Plot a structure's cross section shape for a given medium."""
        mat_index = self.simulation.medium_map[medium]
        plot_params_struct = self.simulation._get_structure_plot_params(
            medium=medium, mat_index=mat_index
        )
        name = medium.name or f"medium[{mat_index}]"
        fig = plotly_shape(shape=shape, plot_params=plot_params_struct, fig=fig, name=name)
        return fig

    @add_fig_if_none
    def plotly_sources(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's sources on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.

        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """

        for source in self.simulation.sources:
            source_plotly = GeometryPlotly(geometry=source)
            fig = source_plotly.plotly(x=x, y=y, z=z, fig=fig, name="sources")
        return fig

    @add_fig_if_none
    def plotly_monitors(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's monitors on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.

        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """

        for monitor in self.simulation.monitors:
            monitor_plotly = GeometryPlotly(geometry=monitor)
            fig = monitor_plotly.plotly(x=x, y=y, z=z, fig=fig, name="monitors")
        return fig

    @add_fig_if_none
    def plotly_pml(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's absorbing boundaries
        on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.


        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """
        normal_axis, _ = self.simulation.parse_xyz_kwargs(x=x, y=y, z=z)
        pml_boxes = self.simulation._make_pml_boxes(normal_axis=normal_axis)
        for pml_box in pml_boxes:
            pml_box_plotly = GeometryPlotly(geometry=pml_box)
            fig = pml_box_plotly.plotly(x=x, y=y, z=z, fig=fig, **plot_params_pml.to_kwargs())
        return fig

    @add_fig_if_none
    def plotly_symmetries(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Plot each of simulation's symmetries on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        fig : plotly.graph_objects.Figure = None
            plotly ``Figure`` to plot on, if not specified, one is created.


        Returns
        -------
        plotly.graph_objects.Figure
            The supplied or created plotly ``Figure``.
        """

        normal_axis, _ = self.simulation.parse_xyz_kwargs(x=x, y=y, z=z)

        for sym_axis, sym_value in enumerate(self.simulation.symmetry):
            if sym_value == 0 or sym_axis == normal_axis:
                continue
            sym_box = self.simulation._make_symmetry_box(sym_axis=sym_axis)
            plot_params = self.simulation._make_symmetry_plot_params(sym_value=sym_value)
            sym_box_plotly = GeometryPlotly(geometry=sym_box)
            fig = sym_box_plotly.plotly(x=x, y=y, z=z, fig=fig, **plot_params.to_kwargs())

        return fig

    @add_fig_if_none
    def _plotly_bounding_box(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        fig: PlotlyFig = None,
    ) -> PlotlyFig:
        """Add simulation bounding box."""
        rmin, rmax = self.simulation.bounds_pml
        sim_box_pml = Box.from_bounds(rmin=rmin, rmax=rmax)
        sim_box_pml_plotly = GeometryPlotly(geometry=sim_box_pml)
        fig = sim_box_pml_plotly.plotly(
            x=x, y=y, z=z, fig=fig, **plot_params_sim_boundary.to_kwargs()
        )
        return fig

    def _plotly_cleanup(
        self,
        fig: PlotlyFig,
        x: float = None,
        y: float = None,
        z: float = None,
    ) -> PlotlyFig:
        """Finish plotting simulation cross section using plotly."""

        normal_axis, pos = self.simulation.parse_xyz_kwargs(x=x, y=y, z=z)

        fig = self._plotly_resize(fig=fig, normal_axis=normal_axis)
        _, (xlabel, ylabel) = self.simulation.pop_axis("xyz", axis=normal_axis)

        fig.update_layout(
            title_text=f'Simulation plotted on {"xyz"[normal_axis]} = {pos:.2f} cross-section',
            title_x=0.5,
            xaxis_title=f"{xlabel} (um)",
            yaxis_title=f"{ylabel} (um)",
            legend_title="Contents",
        )

        fig = self._plotly_clean_labels(fig=fig)

        return fig

    def _plotly_bounds(self, normal_axis: Axis) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """get X, Y limits for plotly figure."""

        rmin, rmax = self.simulation.bounds_pml
        rmin = np.array(rmin)
        rmax = np.array(rmax)
        _, (xmin, ymin) = self.simulation.pop_axis(rmin, axis=normal_axis)
        _, (xmax, ymax) = self.simulation.pop_axis(rmax, axis=normal_axis)
        return (xmin, xmax), (ymin, ymax)

    def _plotly_resize(
        self,
        fig: PlotlyFig,
        normal_axis: Axis,
    ) -> PlotlyFig:
        """Set the lmits and make equal aspect."""

        (xmin, xmax), (ymin, ymax) = self._plotly_bounds(normal_axis=normal_axis)

        fig.update_xaxes(range=[xmin, xmax])
        fig.update_yaxes(range=[ymin, ymax])

        return fig

    @staticmethod
    def _plotly_clean_labels(fig: PlotlyFig) -> PlotlyFig:
        """Remove label entries that show up more than once."""
        seen = []
        for trace in fig["data"]:
            name = trace["name"]
            if name is not None and name not in seen:
                seen.append(name)
            else:
                trace["showlegend"] = False
        return fig



""" Allows users to log in and be authenticated."""
import os
import getpass
import hashlib
import json

import requests

from .config import DEFAULT_CONFIG as Config
from ..log import log, WebError

# maximum attempts for credentials input
MAX_ATTEMPTS = 3

# where we store the credentials locally
CREDENTIAL_FILE = "~/.tidy3d/auth.json"
credential_path = os.path.expanduser(CREDENTIAL_FILE)
credential_dir = os.path.split(credential_path)[0]


def set_authentication_config(email: str, password: str) -> None:
    """Sets the authorization and keys in the config for a for user."""
    os.environ["TIDY3D_USER"] = email
    os.environ["TIDY3D_PASS_HASH"] = password

    url = "/".join([Config.auth_api_endpoint, "auth"])
    headers = {"Application": "TIDY3D"}
    resp = requests.get(url, headers=headers, auth=(email, password))
    access = resp.json()["data"]
    keys = access["user"]

    new_keys = {}
    for key, val in keys.items():
        # does this just create a Title-cased copy of the key for the value?
        new_keys["".join(key[:1].upper() + key[1:])] = val
        new_keys[key] = val

    Config.auth = access["auth"]
    Config.user = new_keys


def encode_password(password: str) -> str:
    """Hash a password."""

    salt = "5ac0e45f46654d70bda109477f10c299"  # TODO: salt should be added server-side
    return hashlib.sha512(password.encode("utf-8") + salt.encode("utf-8")).hexdigest()


# pylint:disable=too-many-branches
def get_credentials() -> None:
    """Tries to log user in from environment variables, then from file, if not working, prompts
    user for login info and saves to file."""

    # if we find credentials in environment variables
    if os.environ.get("TIDY3D_USER") and (
        os.environ.get("TIDY3D_PASS") or os.environ.get("TIDY3D_PASS_HASH")
    ):
        log.debug("Using Tidy3D credentials from environment")
        email = os.environ["TIDY3D_USER"]
        password = os.environ.get("TIDY3D_PASS")
        if password is None:
            password = os.environ.get("TIDY3D_PASS_HASH")
        else:
            password = encode_password(password)
        try:
            set_authentication_config(email, password)
            return

        except Exception:  # pylint:disable=broad-except
            log.info("Error: Failed to log in with environment credentials.")

    # if we find something in the credential path
    if os.path.exists(credential_path):
        log.info("Using Tidy3D credentials from stored file")
        # try to authenticate them
        try:
            with open(credential_path, "r", encoding="utf-8") as fp:
                auth_json = json.load(fp)
            email = auth_json["email"]
            password = auth_json["password"]
            set_authentication_config(email, password)
            return

        except Exception:  # pylint:disable=broad-except
            log.info("Error: Failed to log in with saved credentials.")

    # keep trying to log in
    for counter in range(MAX_ATTEMPTS):

        email = input("enter your email registered at tidy3d: ")
        password = getpass.getpass("enter your password: ")
        # encrypt
        password = encode_password(password)

        try:
            set_authentication_config(email, password)
            break

        except Exception as e:  # pylint:disable=broad-except
            if counter < MAX_ATTEMPTS - 1:
                log.info("Error: Failed to log in with new username and password.")
            else:
                raise WebError("Failed to log in with new username and password.") from e

    # ask to stay logged in
    for _ in range(MAX_ATTEMPTS):

        keep_logged_in = input("Do you want to keep logged in on this machine? ([Y]es / [N]o) ")

        # if user wants to stay logged in
        if keep_logged_in.lower() == "y":

            try:
                if not os.path.exists(credential_dir):
                    os.mkdir(credential_dir)

                auth_json = {"email": email, "password": password}
                with open(credential_path, "w", encoding="utf-8") as fp:
                    json.dump(auth_json, fp)
                return

            except Exception:  # pylint:disable=broad-except
                log.info("Error: Failed to store credentials.")
                return

        # if doesn't want to keep logged in, just return without saving file
        if keep_logged_in.lower() == "n":
            return

        # otherwise, prompt again
        log.info(f"Unknown response: {keep_logged_in}")



"""Import post run visualization app and Simulation plotting through plotly."""

from ...log import Tidy3dImportError

# try to get the plotly packages, otherwise print a helpful error message.
try:
    from jupyter_dash import JupyterDash
    from dash import Dash
    import plotly.graph_objects as go
except ImportError as e:
    raise Tidy3dImportError(
        "Could not import plotly requirements. "
        "Ensure that tidy3d is installed with [plotly] requirements specified. "
        '``pip install "tidy3d-beta[plotly]" or `pip install -e ".[plotly]". '
        "Or, install the dependencies directly with `pip install -r requirements/plotly.txt`"
    ) from e

from .app import SimulationDataApp
from .simulation import SimulationPlotly



"""Defines the structure of the components displayed in the app tabs."""
from abc import ABC, abstractmethod
from dash import dcc

from .utils import PlotlyFig
from ...components.base import Tidy3dBaseModel

"""
How the components work.

1. The app is initialized in `App`.
2. For each UI component in the view:
    a) construct UI component using SimulationData object contents.
    b) call `make_component(app)`, which
        i. generates the layout for the component.
        ii. adds any callback functions to `app` to make it interactive.
        iii. returns the layout.
    c) add the layout from this component to the app.

If a figure is needed, the component can create it with `fig = self.make_figure()`
This will use the internal state to contruct kwargs for a call to `fig = self.plotly(**kwargs)`

"""


class UIComponent(Tidy3dBaseModel, ABC):
    """Base class for a UI component.  Individual data storage wrappers override this."""

    @abstractmethod
    def make_component(self) -> dcc.Tab:
        """Creates the dash component for this montor data."""

    @abstractmethod
    def make_figure(self) -> PlotlyFig:
        """Creates a plotly figure for this component given its current state."""



import numpy as np
import pytest

import tidy3d as td
from tidy3d.constants import fp_eps

from tidy3d.components.grid.mesher import GradedMesher

MESHER = GradedMesher()


def validate_dl_multiple_interval(
    dl_list,
    max_scale,
    max_dl_list,
    len_interval_list,
    is_periodic,
):
    """Validate dl_list"""

    # in each interval
    num_intervals = len(len_interval_list)
    right_dl = np.roll(max_dl_list, shift=-1)
    left_dl = np.roll(max_dl_list, shift=1)
    if not is_periodic:
        right_dl[-1] = max_dl_list[-1]
        left_dl[0] = max_dl_list[0]

    left_dl *= max_scale
    right_dl *= max_scale

    for i in range(num_intervals):
        validate_dl_in_interval(
            dl_list[i], max_scale, left_dl[i], right_dl[i], max_dl_list[i], len_interval_list[i]
        )

    dl_list = np.concatenate(dl_list)
    ratio_f = np.all(dl_list[1:] / dl_list[:-1] <= max_scale + fp_eps)
    ratio_b = np.all(dl_list[1:] / dl_list[:-1] >= 1 / (max_scale + fp_eps))
    assert (ratio_f and ratio_b) == True
    assert np.min(dl_list) >= 0.5 * np.min(max_dl_list)


def validate_dl_in_interval(
    dl_list,
    max_scale,
    left_neighbor_dl,
    right_neighbor_dl,
    max_dl,
    len_interval,
):
    """Validate dl_list"""
    ratio_f = np.all(dl_list[1:] / dl_list[:-1] <= max_scale + fp_eps)
    ratio_b = np.all(dl_list[1:] / dl_list[:-1] >= 1 / (max_scale + fp_eps))
    assert (ratio_f and ratio_b) == True

    left_dl = min(max_dl, left_neighbor_dl)
    right_dl = min(max_dl, right_neighbor_dl)

    assert dl_list[0] <= left_dl + fp_eps
    assert dl_list[-1] <= right_dl + fp_eps
    assert np.max(dl_list) <= max_dl + fp_eps
    assert np.isclose(np.sum(dl_list), len_interval, rtol=fp_eps)


def test_uniform_grid_in_interval():
    """Uniform mesh in an interval"""

    for i in range(100):
        len_interval = 10.0 - np.random.random(1)[0]
        # max_scale = 1, but left_dl != right_dl
        left_dl = np.random.random(1)[0]
        right_dl = np.random.random(1)[0]
        max_dl = np.random.random(1)[0]
        max_scale = 1
        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        assert np.any(dl - dl[0]) == False
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

        # max_scale > 1, but left_dl = right_dl
        left_dl = np.random.random(1)[0]
        right_dl = left_dl
        max_scale = 1 + np.random.random(1)[0]
        max_dl = left_dl
        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        assert np.any(dl - dl[0]) == False
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

        # single pixel
        left_dl = np.random.random(1)[0] + len_interval
        right_dl = np.random.random(1)[0] + len_interval
        max_scale = 1 + np.random.random(1)[0]
        max_dl = left_dl
        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        assert len(dl) == 1
        assert dl[0] == len_interval


def test_asending_grid_in_interval():
    """Nonuniform mesh in an interval from small to large"""

    # # sufficient remaining part, can be inserted
    len_interval = 1.3
    max_scale = 2
    left_dl = 0.3
    right_dl = 1.0
    max_dl = right_dl

    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # remaining part not sufficient to insert, but will not
    # violate max_scale by repearting 1st step
    len_interval = 1.0
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # scaling
    max_scale = 1.1
    len_interval = 0.2 * (1 - max_scale**3) / (1 - max_scale) + 0.14
    left_dl = 0.2
    right_dl = 1.0
    max_dl = right_dl
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # randoms
    for i in range(100):
        max_scale = 1 + np.random.random(1)[0]
        left_dl = np.random.random(1)[0]
        right_dl = 10
        max_dl = 10

        N_step = 1 + np.log(max_dl / left_dl) / np.log(max_scale)
        N_step *= 0.49 + np.random.random(1)[0] * 0.5
        N_step = int(np.floor(N_step))
        len_interval = left_dl * max_scale * (1 - max_scale**N_step) / (1 - max_scale)
        len_interval *= np.random.random(1)[0]

        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

        # opposite direction
        left_dl, right_dl = right_dl, left_dl
        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)


def test_asending_plateau_grid_in_interval():
    """Nonuniform mesh in an interval from small to large to plateau"""

    # # zero pixel for plateau, still asending
    len_interval = 1.0
    max_scale = 2
    left_dl = 0.3
    right_dl = 10
    max_dl = 0.6
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # # sufficient remaining part, can be inserted
    len_interval = 1.9
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # randoms
    for i in range(100):
        max_scale = 1 + np.random.random(1)[0]
        left_dl = np.random.random(1)[0]
        right_dl = 10
        max_dl = 2 + np.random.random(1)[0] * 2

        N_step = 1 + np.log(max_dl / left_dl) / np.log(max_scale)
        N_step = int(np.floor(N_step))
        len_interval = left_dl * max_scale * (1 - max_scale**N_step) / (1 - max_scale)
        len_interval += max_dl * np.random.randint(1, 100)

        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)
        # print(left_dl*max_scale)
        # print(max_dl)
        # print(dl)

        # opposite direction
        left_dl, right_dl = right_dl, left_dl
        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)


def test_asending_plateau_desending_grid_in_interval():
    """Nonuniform mesh in an interval from small to plateau to small"""

    max_scale = 2
    left_dl = 0.1
    right_dl = 0.3
    max_dl = 0.5
    len_interval = 1.51
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    # randoms
    for i in range(100):
        max_scale = 1 + np.random.random(1)[0]
        left_dl = np.random.random(1)[0]
        right_dl = np.random.random(1)[0]
        max_dl = 2 + np.random.random(1)[0] * 2

        N_left_step = 1 + np.log(max_dl / left_dl) / np.log(max_scale)
        N_right_step = 1 + np.log(max_dl / right_dl) / np.log(max_scale)
        N_left_step = int(np.floor(N_left_step))
        N_right_step = int(np.floor(N_right_step))
        len_interval = left_dl * max_scale * (1 - max_scale**N_left_step) / (1 - max_scale)
        len_interval += right_dl * max_scale * (1 - max_scale**N_right_step) / (1 - max_scale)
        len_interval += max_dl * (1 + np.random.randint(1, 100))

        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)


def test_asending_desending_grid_in_interval():
    """Nonuniform mesh in an interval from small to plateau to small"""

    max_scale = 2
    left_dl = 0.1
    right_dl = 0.3
    max_dl = 1
    len_interval = 3.2
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)

    max_scale = 2
    left_dl = 0.3
    right_dl = 0.4
    max_dl = 1
    len_interval = 0.8
    dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
    # print(dl)

    # randoms
    for i in range(100):
        max_scale = 1 + np.random.random(1)[0]
        left_dl = np.random.random(1)[0]
        right_dl = np.random.random(1)[0]
        max_dl = 2 + np.random.random(1)[0] * 2

        N_left_step = 1 + np.log(max_dl / left_dl) / np.log(max_scale)
        N_right_step = 1 + np.log(max_dl / right_dl) / np.log(max_scale)
        N_left_step = int(np.floor(N_left_step))
        N_right_step = int(np.floor(N_right_step))
        len_interval = left_dl * max_scale * (1 - max_scale**N_left_step) / (1 - max_scale)
        len_interval += right_dl * max_scale * (1 - max_scale**N_right_step) / (1 - max_scale)
        len_interval -= max_dl
        len_interval *= np.random.random(1)[0]

        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)


def test_grid_in_interval():
    """Nonuniform mesh in an interval"""

    # randoms
    for i in range(100):
        max_scale = 1 + np.random.random(1)[0]
        left_dl = np.random.randint(1, 10) * np.random.random(1)[0]
        right_dl = np.random.randint(1, 10) * np.random.random(1)[0]
        max_dl = np.random.randint(1, 10) * np.random.random(1)[0]

        len_interval = np.random.randint(1, 100) * np.random.random(1)[0]

        dl = MESHER.make_grid_in_interval(left_dl, right_dl, max_dl, max_scale, len_interval)
        validate_dl_in_interval(dl, max_scale, left_dl, right_dl, max_dl, len_interval)


def test_grid_analytic_refinement():

    max_dl_list = np.array([0.5, 0.5, 0.4, 0.1, 0.4])
    len_interval_list = np.array([2.0, 0.5, 0.2, 0.1, 0.3])
    max_scale = 1.5
    periodic = True
    left_dl, right_dl = MESHER.grid_multiple_interval_analy_refinement(
        max_dl_list, len_interval_list, max_scale, periodic
    )
    assert np.all(np.isclose(left_dl[1:], right_dl[:-1])) == True


def test_grid_refinement():

    max_dl_list = np.array([0.5, 0.4, 0.1, 0.4])
    len_interval_list = np.array([0.5, 1.2, 0.1, 1.3])
    max_scale = 1.5
    is_periodic = False
    dl_list = MESHER.make_grid_multiple_intervals(
        max_dl_list, len_interval_list, max_scale, is_periodic
    )
    # print(np.min(np.concatenate(dl_list))/np.min(max_dl_list))

    validate_dl_multiple_interval(
        dl_list,
        max_scale,
        max_dl_list,
        len_interval_list,
        is_periodic,
    )

    num_intervals = 100
    max_shrink = 1
    for i in range(50):
        max_dl_list = np.random.random(num_intervals)
        len_interval_list = np.random.random(num_intervals) * 10
        too_short_ind = len_interval_list < max_dl_list
        len_interval_list[too_short_ind] = max_dl_list[too_short_ind] * (1 + np.random.random(1)[0])
        max_scale = 1.1
        is_periodic = True
        dl_list = MESHER.make_grid_multiple_intervals(
            max_dl_list, len_interval_list, max_scale, is_periodic
        )
        shrink_local = np.min(np.concatenate(dl_list)) / np.min(max_dl_list)
        if shrink_local < max_shrink:
            max_shrink = shrink_local
        validate_dl_multiple_interval(
            dl_list,
            max_scale,
            max_dl_list,
            len_interval_list,
            is_periodic,
        )


WAVELENGTH = 2.9

BOX1 = td.Structure(
    geometry=td.Box(center=(0, 0, 0), size=(2, 2, 2)), medium=td.Medium(permittivity=9)
)
# covers BOX1 along x and y but not z, smaller permittivity
BOX2 = td.Structure(
    geometry=td.Box(center=(0, 0, 0), size=(200, 200, 1)), medium=td.Medium(permittivity=4)
)
# covers BOX1 along x only, smaller permittivity
BOX3 = td.Structure(
    geometry=td.Box(center=(0, 1.5, 0), size=(200, 4, 1)), medium=td.Medium(permittivity=4)
)
# fully covers one edge of BOX1
BOX4 = td.Structure(
    geometry=td.Box(center=(0, 1.01, 0), size=(200, 0.2, 2)), medium=td.Medium(permittivity=2)
)
# box made out of gold
GOLD = td.material_library["Au"]["JohnsonChristy1972"]
BOX5 = td.Structure(geometry=td.Box(center=(0, 0, 0), size=(1, 1, 0.1)), medium=GOLD)
# fully covers BOX5, regular dielectric
BOX6 = td.Structure(
    geometry=td.Box(center=(0, 0, 0), size=(1, 1, 0.2)), medium=td.Medium(permittivity=2)
)


def test_mesh_structure_covers():
    # Test that the BOX2 permittivity is used along z in the region where it fully covers BOX1
    sim = td.Simulation(
        size=(3, 3, 3),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX1, BOX2],
    )
    sizes = sim.grid.sizes.to_list[2]
    assert sizes[len(sizes) // 2] > 0.1


def test_mesh_structure_partially_covers():
    # Test that the BOX3 permittivity is not used along z as it doesn't fully cover BOX1
    sim = td.Simulation(
        size=(3, 3, 3),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX1, BOX3],
    )
    sizes = sim.grid.sizes.to_list[2]
    assert sizes[len(sizes) // 2] < 0.1


def test_mesh_structure_covers_boundary():
    # Test that there is no grid boundary along y at the BOX1 right side covered by BOX4
    sim = td.Simulation(
        size=(3, 3, 3),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX1, BOX4],
    )
    boundaries = sim.grid.boundaries.to_list[1]
    assert 1.0 not in boundaries


def test_mesh_high_index_background():
    # Test high-index background medium
    sim = td.Simulation(
        size=(3, 3, 6),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX1, BOX2],
        medium=td.Medium(permittivity=5**2),
    )
    sizes = sim.grid.sizes.to_list[2]
    assert sizes[0] < WAVELENGTH / 50


def test_mesh_high_index_background_override():
    # Test high-index background with override box
    sim = td.Simulation(
        size=(3, 3, 6),
        grid_spec=td.GridSpec.auto(
            wavelength=WAVELENGTH,
            override_structures=[
                td.Structure(
                    geometry=td.Box(size=(td.inf, td.inf, td.inf)),
                    medium=td.Medium(permittivity=1),
                ),
                BOX1,
                BOX2,
            ],
        ),
        run_time=1e-13,
        structures=[BOX1, BOX2],
        medium=td.Medium(permittivity=5**2),
    )
    sizes = sim.grid.sizes.to_list[2]
    assert np.isclose(sizes[0], WAVELENGTH / 10)


def test_mesh_gold_slab():
    # Test meshing of a slab with large negative permittivity
    gold_step = WAVELENGTH / 10 / np.sqrt(np.abs(GOLD.eps_model(td.C_0 / WAVELENGTH).real))

    sim = td.Simulation(
        size=(3, 3, 6),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX5],
    )
    sizes = sim.grid.sizes.to_list[2]
    assert np.amin(sizes) < gold_step

    # Test that the minimum step is overridden if the gold slab is covered
    # This includes checking that only one of the dielectric box boundaries along z is added
    sim = td.Simulation(
        size=(3, 3, 6),
        grid_spec=td.GridSpec.auto(wavelength=WAVELENGTH),
        run_time=1e-13,
        structures=[BOX5, BOX6],
    )
    sizes = sim.grid.sizes.to_list[2]
    assert np.amin(sizes) > BOX6.geometry.size[2]


@pytest.mark.timeout(3.0)
def test_mesher_timeout():
    """Test that the mesh generation is fast."""
    np.random.seed(4)
    num_boxes = 500
    box_scale = 5
    sim_size = 5
    n_max = 5
    mediums = [td.Medium(permittivity=n**2) for n in (1 + (n_max - 1) * np.random.rand(100))]

    boxes = []
    for i in range(num_boxes):
        center = sim_size * (np.random.rand(3) - 0.5)
        center[0] = 0
        size = np.abs(box_scale * np.random.randn(3))
        n = 1 + (n_max - 1) * np.random.rand(1)
        box = td.Structure(
            geometry=td.Box(center=center.tolist(), size=size.tolist()),
            medium=mediums[np.random.randint(0, 100)],
        )
        boxes.append(box)

    sim = td.Simulation(
        size=(sim_size,) * 3,
        grid_spec=td.GridSpec.auto(min_steps_per_wvl=6),
        run_time=1e-13,
        structures=boxes,
        sources=[
            td.PointDipole(
                source_time=td.GaussianPulse(freq0=2e14, fwidth=1e13),
                size=(0, 0, 0),
                polarization="Ex",
            )
        ],
    )

    grid = sim.grid



"""test slanted polyslab can be correctly setup and visualized. """
from typing import Dict
import pytest
import numpy as np
import pydantic
from shapely.geometry import Polygon, Point

import tidy3d as td
from tidy3d.log import ValidationError, SetupError

np.random.seed(4)


def setup_polyslab(vertices, dilation, angle, bounds):
    """Setup slanted polyslab"""
    s = td.PolySlab(
        vertices=vertices,
        slab_bounds=bounds,
        axis=2,
        dilation=dilation,
        sidewall_angle=angle,
    )
    return s


def minimal_edge_length(vertices):
    """compute the minimal edge length in a polygon"""
    vs = vertices.T.copy()
    vsp = np.roll(vs.copy(), axis=-1, shift=-1)
    edge_length = np.linalg.norm(vsp - vs, axis=0)
    return np.min(edge_length)


def convert_valid_polygon(vertices):
    """Given vertices that might have intersecting edges, converted to
    vertices of a valid polygon
    """
    poly = Polygon(vertices).buffer(0)  # make sure no intersecting edges
    if type(poly) is not Polygon:
        poly = poly.geoms[0]

    vertices_n = np.array(poly.exterior.coords[:])
    return vertices_n


def validate_poly_bound(poly):
    """validate bound based polyslab's base and top polygon"""
    xmin1, ymin1 = np.amin(poly.base_polygon, axis=0)
    xmax1, ymax1 = np.amax(poly.base_polygon, axis=0)

    xmin2, ymin2 = np.amin(poly.top_polygon, axis=0)
    xmax2, ymax2 = np.amax(poly.top_polygon, axis=0)

    xmin, ymin = min(xmin1, xmin2), min(ymin1, ymin2)
    xmax, ymax = max(xmax1, xmax2), max(ymax1, ymax2)

    bound_tidy = poly.bounds
    assert bound_tidy[0][0] <= xmin
    assert bound_tidy[0][1] <= ymin
    assert bound_tidy[1][0] >= xmax
    assert bound_tidy[1][1] >= ymax


# default values
bounds = (0, 0.5)
dilation = 0.0
angle = 0


def test_remove_duplicate():
    """
    Make sure redundant neighboring vertices are removed
    """
    vertices = np.random.random((10, 2))
    vertices[0, :] = vertices[9, :]
    vertices[1, :] = vertices[0, :]
    vertices[5, :] = vertices[6, :]

    vertices = td.PolySlab._remove_duplicate_vertices(vertices)
    assert vertices.shape[0] == 7


def test_valid_polygon():
    """No intersecting edges"""

    # area = 0
    vertices = ((0, 0), (1, 0), (2, 0))
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # only two points
    vertices = ((0, 0), (1, 0), (1, 0))
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # intersecting edges
    vertices = ((0, 0), (1, 0), (1, 1), (0, 1), (0.5, -1))

    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)


def test_crossing_square():
    """
    Vertices crossing detection for a simple square
    """
    vertices = ((0, 0), (1, 0), (1, -1), (0, -1))
    dilation = 0.0
    angle = np.pi / 4
    s = setup_polyslab(vertices, dilation, angle, bounds)

    # dilation too significant
    dilation = -1.1
    angle = 0
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # angle too large
    dilation = 0
    angle = np.pi / 3
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # combines both
    dilation = -0.1
    angle = np.pi / 4
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)


def test_max_erosion_polygon():
    """
    Maximal erosion distance validation
    """
    N = 10  # number of vertices
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * 10)

        dilation = 0
        angle = 0
        bounds = (0, 0.5)
        s = setup_polyslab(vertices, dilation, angle, bounds)

        # compute maximal allowed erosion distance
        _, max_dist = s._crossing_detection(s.base_polygon, -100)

        # verify it is indeed maximal allowed
        dilation = -max_dist + 1e-10
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        assert np.isclose(minimal_edge_length(s.base_polygon), 0, atol=1e-4)

        # verify it is indeed maximal allowed
        dilation = 0.0
        bounds = (0, max_dist - 1e-10)
        angle = np.pi / 4

        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        assert np.isclose(minimal_edge_length(s.top_polygon), 0, atol=1e-4)


def test_shift_height():
    """Make sure a list of height where the plane will intersect with the vertices
    works properly
    """
    N = 10  # number of vertices
    Lx = 10.0
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, max_dist * 0.99)
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue

        for axis in (0, 1):
            position = np.random.random(1)[0] * Lx - Lx / 2
            height = s._find_intersecting_height(position, axis)
            for h in height:
                bounds = (0, h)
                s = setup_polyslab(vertices, dilation, angle, bounds)
                diff = s.top_polygon[:, axis] - position
                assert np.any(np.isclose(diff, 0)) == True


def test_intersection_with_inside():
    """Make sure intersection produces the same result as inside"""

    N = 10  # number of vertices
    Lx = 10  # maximal length in x,y direction
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        vertices = np.array(vertices)
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)

        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, (max_dist * 0.95))
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue

        ### side x
        xp = np.random.random(1)[0] * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(x=xp)

        for i in range(len(yp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp, yp[i], zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(yp[i], zp[j])):
                        res_inter = True
                # if res_inter != res_inside:
                #     print(repr(vertices))
                #     print(repr(s.base_polygon))
                #     print(bounds)
                #     print(xp, yp[i], zp[j])
                assert res_inter == res_inside

        ### side y
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(1)[0] * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(y=yp)

        for i in range(len(xp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp[i], yp, zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], zp[j])):
                        res_inter = True
                assert res_inter == res_inside

        ### norm z
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(1)[0] * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(z=zp)

        for i in range(len(xp)):
            for j in range(len(yp)):
                # inside
                res_inside = s.inside(xp[i], yp[j], zp)
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], yp[j])):
                        res_inter = True
                assert res_inter == res_inside


def test_intersection_with_inside_negative_angle():
    """Make sure intersection produces the same result as inside
    for slantwall angle < 0
    """

    N = 10  # number of vertices
    Lx = 10  # maximal length in x,y direction
    max_dist = 5
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        vertices = np.array(vertices)

        dilation = 0.0
        bounds = (0, (max_dist * 0.95))
        angle = -np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue

        ### side x
        xp = np.random.random(1)[0] * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(x=xp)

        for i in range(len(yp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp, yp[i], zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(yp[i], zp[j])):
                        res_inter = True
                # if res_inter != res_inside:
                #     print("============================")
                #     print(repr(vertices))
                #     print(repr(s.base_polygon))
                #     print(bounds)
                #     print(xp, yp[i], zp[j])
                #     print('len = ', len(shape_intersect))
                #     for shape in shape_intersect:
                #         print(list(shape.exterior.coords))
                #         print(shape.covers(Point(yp[i],zp[j])))

                #     yp = np.linspace(0,10,200)
                #     zp = np.linspace(0.,bounds[1],100)
                #     contain = np.zeros((len(yp),len(zp)),dtype=bool)
                #     for ii in range(len(yp)):
                #         for jj in range(len(zp)):
                #             contain[ii][jj]=s.inside(xp,yp[ii],zp[jj])

                #     intersect = np.zeros((len(yp),len(zp)),dtype=bool)
                #     for ii in range(len(yp)):
                #         for jj in range(len(zp)):
                #             for shape in shape_intersect:
                #                 if shape.covers(Point(yp[ii],zp[jj])):
                #                     intersect[i][j] = True
                #     import matplotlib.pyplot as plt
                #     fig, ax = plt.subplots(1, 2, constrained_layout=True)
                #     ax[0].imshow(contain.T,origin='lower',extent=[yp[0],yp[-1],zp[0],zp[-1]],aspect='auto')
                #     ax[1].imshow(intersect.T,origin='lower',extent=[yp[0],yp[-1],zp[0],zp[-1]],aspect='auto')
                #     plt.show()
                assert res_inter == res_inside

        ### side y
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(1)[0] * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(y=yp)

        for i in range(len(xp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp[i], yp, zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], zp[j])):
                        res_inter = True
                assert res_inter == res_inside

        ### norm z
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(1)[0] * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(z=zp)

        for i in range(len(xp)):
            for j in range(len(yp)):
                # inside
                res_inside = s.inside(xp[i], yp[j], zp)
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], yp[j])):
                        res_inter = True
                assert res_inter == res_inside


def test_bound():
    """
    Make sure bound works, even though it might not be tight.
    """
    N = 10  # number of vertices
    Lx = 10  # maximal length in x,y direction
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        vertices = np.array(vertices)  # .astype("float32")

        ### positive dilation
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        _, max_dist = s._crossing_detection(s.base_polygon, 100)
        # verify it is indeed maximal allowed
        dilation = 1
        if max_dist is not None:
            dilation = max_dist - 1e-10
        bounds = (0, 1)
        angle = 0.0
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        validate_poly_bound(s)

        ## sidewall
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, (max_dist * 0.95))
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        validate_poly_bound(s)



""" imports interfaces for interacting with server """
import sys

from .webapi import run, upload, get_info, start, monitor, delete, download, load, estimate_cost
from .webapi import get_tasks, delete_old, download_json, download_log, load_simulation
from .container import Job, Batch, BatchData
from .auth import get_credentials



"""External plugins that have tidy3d as dependency and add functionality."""

from .dispersion.fit import DispersionFitter
from .dispersion.fit_web import StableDispersionFitter, AdvancedFitterParam
from .mode.mode_solver import ModeSolver, ModeSolverData
from .smatrix.smatrix import ComponentModeler, Port



"""Logging and error-handling for Tidy3d."""
import logging
from rich.logging import RichHandler

# TODO: more logging features (to file, etc).

FORMAT = "%(message)s"

DEFAULT_LEVEL = "INFO"

logging.basicConfig(level=DEFAULT_LEVEL, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

# maps level string to level integer for python logging package
LEVEL_MAP = {
    "error": 40,
    "warning": 30,
    "info": 20,
    "debug": 10,
}

# importable logger
log = logging.getLogger("rich")


""" Tidy3d custom exceptions """


class Tidy3dError(Exception):
    """Any error in tidy3d"""

    def __init__(self, message: str = None):
        """Log just the error message and then raise the Exception."""
        log.error(message)
        super().__init__(self, message)


class ConfigError(Tidy3dError):
    """Error when configuring Tidy3d."""


class Tidy3dKeyError(Tidy3dError):
    """Could not find a key in a Tidy3d dictionary."""


class ValidationError(Tidy3dError):
    """Error when constructing Tidy3d components."""


class SetupError(Tidy3dError):
    """Error regarding the setup of the components (outside of domains, etc)."""


class FileError(Tidy3dError):
    """Error reading or writing to file."""


class WebError(Tidy3dError):
    """Error with the webAPI."""


class AuthenticationError(Tidy3dError):
    """Error authenticating a user through webapi webAPI."""


class DataError(Tidy3dError):
    """Error accessing data."""


class Tidy3dImportError(Tidy3dError):
    """Error importing a package needed for tidy3d."""


""" Logging functions """


def _get_level_int(level: str) -> int:
    """Get the integer corresponding to the level string."""
    level = level.lower()
    if level not in LEVEL_MAP:
        raise ConfigError(
            f"logging level {level} not supported, must be in {list(LEVEL_MAP.keys())}."
        )
    return LEVEL_MAP[level]


def set_logging_level(level: str = DEFAULT_LEVEL.lower()) -> None:
    """Set tidy3d logging level priority.

    Parameters
    ----------
    level : str = 'info'
        The lowest priority level of logging messages to display.
        One of ``{'debug', 'info', 'warning', 'error'}`` (listed in increasing priority).

    Example
    -------
    >>> log.debug('this message should not appear (default logging level = INFO')
    >>> set_logging_level('debug')
    >>> log.debug('this message should appear now')
    """

    level_int = _get_level_int(level)
    log.setLevel(level_int)


def set_logging_file(fname: str, filemode="w", level=DEFAULT_LEVEL.lower()):
    """Set a file to write log to, independently from the stdout and stderr
    output chosen using :meth:`set_logging_level`.

    Parameters
    ----------
    fname : str
        Path to file to direct the output to.
    filemode : str = 'w'
        'w' or 'a', defining if the file should be overwritten or appended.
    level : str = 'info'
        One of 'debug', 'info', 'warning', 'error', 'critical'. This is
        set for the file independently of the console output level set by
        :meth:`set_logging_level`.

    Example
    -------
    >>> set_logging_file('tidy3d_log.log')
    >>> log.warning('this warning will appear in the tidy3d_log.log')
    """

    file_handler = logging.FileHandler(fname, filemode)
    level_int = _get_level_int(level)
    file_handler.setLevel(level_int)
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    file_handler.setFormatter(formatter)
    log.addHandler(file_handler)



"""Sets the configuration of the script, can be changed with `td.config.config_name = new_val`."""

import pydantic as pd
from typing_extensions import Literal

from .log import set_logging_level, DEFAULT_LEVEL


class Tidy3dConfig(pd.BaseModel):
    """configuration of tidy3d"""

    class Config:
        """Config of the config."""

        arbitrary_types_allowed = False
        validate_all = True
        extra = "forbid"
        validate_assignment = True
        allow_population_by_field_name = True
        frozen = False

    logging_level: Literal["debug", "info", "warning", "error"] = pd.Field(
        DEFAULT_LEVEL.lower(),
        title="Logging Level",
        description="The lowest level of logging output that will be displayed. "
        'Can be "debug", "info", "warning", "error".',
    )

    @pd.validator("logging_level", always=True)
    def _set_logging_level(cls, val):
        """Set the logging level if logging_level is changed."""
        set_logging_level(val)
        return val


# instance of the config that can be modified.
config = Tidy3dConfig()
